from . cimport everything as cqlat
from .geometry cimport Geometry

cdef class SelectedShufflePlan:

    cdef cqlat.SelectedShufflePlan xx
    cdef public list psel_src_list
    cdef public list psel_dst_list
    cdef public list fsel_src_list
    cdef public list fsel_dst_list
    cdef public list geo_src_list
    cdef public list geo_dst_list

cdef class PointsSelection:

    cdef cqlat.PointsSelection xx
    cdef readonly cqlat.Long cdata
    cdef readonly int view_count

cdef class FieldSelection:

    cdef cqlat.FieldSelection xx
    cdef readonly cqlat.Long cdata
    cdef readonly int view_count



__all__ = [
        'prop_spatial_smear',
        ]

import numpy as np

class q:
    from qlat_utils import (
            timer,
            get_chunk_list,
            CoordinateD,
            mk_cache,
            )
    from qlat.c import (
            GaugeField,
            Prop,
            FermionField4d,
            PointsSelection,
            SelectedShufflePlan,
            prop_spatial_smear_no_comm,
            mk_ff_list_from_prop,
            mk_prop_from_ff_list,
            )

cache_spatial_smear_chunk_plan = q.mk_cache("spatial_smear_chunk_plan")

@q.timer(is_flops=True)
def prop_spatial_smear(ff_list, gf, coef, step, mom=None, *, chunk_size=None):
    """
    Perform spatial smear for `ff_list`, a list of `FermionField4d`.
    Return new `ff_list` after smearing.
    Original `ff_list` should not be modified.
    #
    `ff_list` can also be a `Prop`.
    In this case, the function will also return a `Prop`.
    #
    get_gf_ape = run_gf_ape(job_tag, get_gf)
    gf = get_gf_ape()
    #
    coef = get_param(job_tag, "prop_smear_coef")
    step = get_param(job_tag, "prop_smear_step")
    #
    job_tag = "64I"
    set_param(job_tag, "gf_ape_smear_coef")(0.5)
    set_param(job_tag, "gf_ape_smear_step")(30)
    set_param(job_tag, "prop_smear_coef")(0.9375)
    set_param(job_tag, "prop_smear_step")(54)
    """
    assert isinstance(gf, q.GaugeField)
    geo = gf.geo
    #
    if isinstance(ff_list, q.FermionField4d):
        ff = ff_list
        [ ff_p, ] = prop_spatial_smear([ ff, ], gf, coef, step, mom, chunk_size=chunk_size)
        return 0, ff_p
    if isinstance(ff_list, q.Prop):
        ff = ff_list
        [ ff_p, ] = prop_spatial_smear([ ff, ], gf, coef, step, mom, chunk_size=chunk_size)
        return 0, ff_p
    #
    assert isinstance(ff_list, list)
    if len(ff_list) == 0:
        return 0, []
    #
    assert isinstance(coef, float)
    assert isinstance(step, int)
    assert step >= 0
    if mom is None:
        mom = q.CoordinateD()
    assert isinstance(mom, q.CoordinateD)
    #
    is_ff_type_prop = isinstance(ff_list[0], q.Prop)
    if is_ff_type_prop:
        for ff in ff_list:
            assert isinstance(ff, q.Prop)
    else:
        for ff in ff_list:
            assert isinstance(ff, q.FermionField4d)
    #
    for ff in ff_list:
        assert geo == ff.geo
    #
    if step == 0:
        return 0, [ ff.copy() for ff in ff_list ]
    #
    if chunk_size is None:
        if is_ff_type_prop:
            chunk_size = 1
        else:
            chunk_size = 12
    #
    chunk_ff_list_list = q.get_chunk_list(ff_list, chunk_size=chunk_size)
    #
    ss_ff_list = []
    #
    plan = dict()
    #
    for chunk_ff_list in chunk_ff_list_list:
        if is_ff_type_prop:
            num_field = len(chunk_ff_list) * 12
        else:
            num_field = len(chunk_ff_list)
        if plan.get("num_field") != num_field:
            plan = get_prop_spatial_smear_chunk_plan(gf, num_field)
        assert plan["num_field"] == num_field
        if is_ff_type_prop:
            chunk_p_list = chunk_ff_list
            chunk_ff_list = []
            for p in chunk_p_list:
                chunk_ff_list += q.mk_ff_list_from_prop(p)
        assert len(chunk_ff_list) == num_field
        ss_chunk_ff_list = prop_spatial_smear_chunk(chunk_ff_list, plan, coef, step, mom)
        assert len(ss_chunk_ff_list) == num_field
        if is_ff_type_prop:
            ss_chunk_p_list = []
            for p_ff_list in q.get_chunk_list(ss_chunk_ff_list, chunk_size=12):
                assert len(p_ff_list) == 12
                ss_chunk_p_list.append(q.mk_prop_from_ff_list(p_ff_list))
            ss_chunk_ff_list = ss_chunk_p_list
        ss_ff_list += ss_chunk_ff_list
    #
    n_avg = 6
    num_field = len(ff_list)
    flops_per_step = geo.local_volume * num_field * 4 * n_avg * (3 * (3 * 6 + 2 * 2))
    flops = flops_per_step * step
    if is_ff_type_prop:
        flops *= 12
    #
    return flops, ss_ff_list

@q.timer
def get_prop_spatial_smear_chunk_plan(gf, num_field):
    """
    Only cache one plan.
    """
    key = "plan"
    if key in cache_spatial_smear_chunk_plan:
        plan = cache_spatial_smear_chunk_plan[key]
        b = True
        b = b and (gf is plan["gf"])
        b = b and (num_field == plan["num_field"])
        if b:
            return plan
    plan = prop_spatial_smear_chunk_planner(gf, num_field)
    cache_spatial_smear_chunk_plan[key] = plan
    return plan

@q.timer
def prop_spatial_smear_chunk_planner(gf, num_field):
    """
    return plan
    #
    plan = dict(s_gf_list=s_gf_list, ssp=ssp)
    """
    geo = gf.geo
    total_site = geo.total_site
    psel = q.PointsSelection(geo)
    psel_list = [ psel.copy() for i in range(num_field) ]
    geo_list = [ geo.copy() for i in range(num_field) ]
    #
    ssp_gf = q.SelectedShufflePlan("dist_t_slice_from_l", psel, geo, num_field)
    ssp = q.SelectedShufflePlan("t_slice_from_l", psel_list, geo_list)
    #
    s_gf_list = ssp_gf.shuffle_sp_list(q.GaugeField, [ gf, ])
    #
    n_avg = 6
    flops_per_step = geo.local_volume * num_field * 4 * n_avg * (3 * (3 * 6 + 2 * 2))
    #
    plan = dict(
            gf=gf,
            num_field=num_field,
            s_gf_list=s_gf_list,
            ssp=ssp,
            flops_per_step=flops_per_step,
            )
    return plan

@q.timer(is_flops=True)
def prop_spatial_smear_chunk(ff_list, plan, coef, step, mom):
    """
    return ss_ff_list
    which is the smeared FermionField4d fields.
    """
    num_field = plan['num_field']
    s_gf_list = plan['s_gf_list']
    ssp = plan['ssp']
    flops_per_step = plan['flops_per_step']
    #
    assert len(ff_list) == num_field
    #
    s_ff_list = ssp.shuffle_sp_list(q.FermionField4d, ff_list)
    #
    id_node_set = set()
    s_ff_mask_arr = np.zeros(len(s_ff_list), dtype=np.int8)
    for s_gf in s_gf_list:
        id_node = s_gf.geo.id_node
        assert id_node not in id_node_set
        id_node_set.add(id_node)
        sub_s_ff_list = []
        for idx, s_ff in enumerate(s_ff_list):
            if id_node == s_ff.geo.id_node:
                assert s_ff_mask_arr[idx] == 0
                s_ff_mask_arr[idx] = 1
                sub_s_ff_list.append(s_ff)
        q.prop_spatial_smear_no_comm(sub_s_ff_list, s_gf, coef, step, mom)
    assert np.all(s_ff_mask_arr == 1)
    #
    ss_ff_list = ssp.shuffle_sp_list(q.FermionField4d, s_ff_list, is_reverse=True)
    #
    flops = flops_per_step * step
    #
    return flops, ss_ff_list



import sys
from pathlib import Path
from Cython.Tempita import sub

"""
Link to some docs
https://github.com/cython/cython/blob/master/Cython/Tempita/_tempita.py
As of writing the version is
https://github.com/cython/cython/blob/023d4af351042a3b6241dbe06cbb003b3ce1fb58/Cython/Tempita/_tempita.py
"""

template = Path(sys.argv[1]).read_text('utf8')
output = sub(template)
Path(sys.argv[2]).write_text(output, 'utf8')



# cython: binding=True, embedsignature=True, c_string_type=unicode, c_string_encoding=utf8

from qlat_utils.all cimport *
from . cimport everything as cc
from .field_base cimport FieldBase, SelectedFieldBase
from .field_selection cimport FieldSelection
from .field_types cimport FieldChar
from .selected_field_types cimport SelectedFieldChar

import qlat_utils as q

cache_fields_io = q.mk_cache("fields_io")

cdef class ShuffledFieldsWriter:

    # self.cdata

    def __cinit__(self):
        self.cdata = <cc.Long>&(self.xx)

    def __init__(self, str path not None, Coordinate new_size_node not None, cc.Bool is_append=False):
        self.xx.init(path, new_size_node.xx, is_append)

    def __del__(self):
        self.close()

    def close(self):
        self.xx.close()
        cache_fields_io.pop(id(self), None)

    def path(self):
        return self.xx.path

    def new_size_node(self):
        cdef Coordinate x = Coordinate()
        x.xx = self.xx.new_size_node
        return x

    def get_cache_sbs(self, FieldSelection fsel):
        if fsel is None:
            cache_fields_io.pop(id(self), None)
            return
        if id(self) in cache_fields_io:
            c_fsel, c_sbs = cache_fields_io[id(self)]
            if fsel is c_fsel:
                return c_sbs
        sbs = ShuffledBitSet(fsel, self.new_size_node())
        assert sbs.xx.fsel.n_elems == fsel.xx.n_elems
        assert cc.is_matching_fsel(sbs.xx.fsel, fsel.xx)
        cache_fields_io[id(self)] = (fsel, sbs,)
        return sbs

    def list(self):
        return cc.list_fields(self.xx)

    def has(self, const cc.std_string& fn):
        return cc.does_file_exist_sync_node(self.xx, fn)

    def __contains__(self, str fn):
        return self.has(fn)

    def flush(self):
        return cc.flush(self.xx)

    def write(self, str fn, obj):
        if isinstance(obj, (FieldBase, SelectedFieldBase,)):
            return obj.write_sfw_direct(self, fn)
        else:
            raise Exception("ShuffledFieldsWriter.write")

## --------------

cdef class ShuffledFieldsReader:

    # self.cdata

    def __cinit__(self):
        self.cdata = <cc.Long>&(self.xx)

    def __init__(self, str path not None, Coordinate new_size_node=None):
        if new_size_node is None:
            new_size_node = Coordinate()
        self.xx.init(path, new_size_node.xx)

    def __del__(self):
        self.close()

    def close(self):
        self.xx.close()
        cache_fields_io.pop(id(self), None)

    def path(self):
        return self.xx.path

    def new_size_node(self):
        cdef Coordinate x = Coordinate()
        x.xx = self.xx.new_size_node
        return x

    def get_cache_sbs(self, FieldSelection fsel):
        if fsel is None:
            cache_fields_io.pop(id(self), None)
            return
        if id(self) in cache_fields_io:
            c_fsel, c_sbs = cache_fields_io[id(self)]
            if fsel is c_fsel:
                return c_sbs
        sbs = ShuffledBitSet(fsel, self.new_size_node())
        assert sbs.xx.fsel.n_elems == fsel.xx.n_elems
        assert cc.is_matching_fsel(sbs.xx.fsel, fsel.xx)
        cache_fields_io[id(self)] = (fsel, sbs,)
        return sbs

    def list(self):
        return cc.list_fields(self.xx)

    def has_duplicates(self):
        return cc.has_duplicates(self.xx)

    def has(self, const cc.std_string& fn):
        return cc.does_file_exist_sync_node(self.xx, fn)

    def __contains__(self, str fn):
        return self.has(fn)

    def is_sparse_field(self, cc.std_string& fn):
        """
        return True if `fn in self` and is sparse field.
        """
        return cc.is_sparse_field_sync_node(self.xx, fn)

    def read_as_char(self, cc.std_string& fn):
        """
        return SelectedFieldChar or FieldChar.
        return None if `fn not in self`.
        """
        has_field = self.has(fn)
        if not has_field:
            return None
        is_sparse = self.is_sparse_field(fn)
        if is_sparse:
            obj = SelectedFieldChar(None)
        else:
            obj = FieldChar()
        total_bytes = obj.read_sfr_direct(self, fn)
        assert total_bytes > 0
        return obj

    def read(self, str fn, obj):
        """
        Can also read SelectedField obj with obj.fsel is None
        After reading, obj.fsel will be properly loaded.
        """
        if isinstance(obj, (FieldBase, SelectedFieldBase,)):
            return obj.read_sfr_direct(self, fn)
        else:
            raise Exception("ShuffledFieldsReader.read")

## --------------

cdef class ShuffledBitSet:

    def __cinit__(self):
        self.cdata = <cc.Long>&(self.xx)

    def __init__(self, FieldSelection fsel not None, Coordinate new_size_node not None):
        self.xx = cc.mk_shuffled_bitset(fsel.xx, new_size_node.xx)

## --------------

@q.timer
def open_fields(str path, str mode, Coordinate new_size_node=None):
    """
    path can be the folder path or the 'geon-info.txt' path
    """
    if path[-14:] == "/geon-info.txt":
        path = path[:-14]
    if mode == "r":
        return ShuffledFieldsReader(path, new_size_node)
    elif mode == "w":
        assert new_size_node is not None
        return ShuffledFieldsWriter(path, new_size_node)
    elif mode == "a":
        if new_size_node is None:
            new_size_node = Coordinate()
        return ShuffledFieldsWriter(path, new_size_node, True)
    else:
        raise Exception("open_fields")

@q.timer
def list_fields(str path, Coordinate new_size_node=None):
    cdef ShuffledFieldsReader sfr = open_fields(path, "r", new_size_node)
    cdef list fns = sfr.list()
    sfr.close()
    return fns

@q.timer
def fields_build_index(str path, Coordinate new_size_node=None):
    cdef ShuffledFieldsReader sfr = open_fields(path, "r", new_size_node)
    cc.fields_build_index(sfr.xx)
    sfr.close()

@q.timer
def fields_has_duplicates(str path, Coordinate new_size_node=None):
    cdef ShuffledFieldsReader sfr = open_fields(path, "r", new_size_node)
    ret = sfr.has_duplicates();
    sfr.close()
    return ret

@q.timer
def properly_truncate_fields(str path, cc.Bool is_check_all=False, cc.Bool is_only_check=False, Coordinate new_size_node=None):
    if path[-14:] == "/geon-info.txt":
        path = path[:-14]
    if new_size_node is None:
        new_size_node = Coordinate()
    return cc.properly_truncate_fields_sync_node(path, is_check_all, is_only_check, new_size_node.xx)

@q.timer
def truncate_fields(str path, list fns_keep, Coordinate new_size_node=None):
    """
    fns_keep is the list of fields that need to keep
    fns_keep needs to be in the same order as the data is stored in path
    all fns_keep must be already in the path
    fns_keep can be empty list
    """
    if path[-14:] == "/geon-info.txt":
        path = path[:-14]
    if new_size_node is None:
        new_size_node = Coordinate()
    ret = cc.truncate_fields_sync_node(path, fns_keep, new_size_node.xx)
    if ret != 0:
        raise Exception(f"truncate_fields: error {ret}")

@q.timer
def check_fields(str path, cc.Bool is_check_all=True, Coordinate new_size_node=None):
    """
    return list of field that is stored successful
    """
    if path[-14:] == "/geon-info.txt":
        path = path[:-14]
    is_only_check = True
    return properly_truncate_fields(path, is_check_all, is_only_check, new_size_node)

@q.timer
def check_compressed_eigen_vectors(str path):
    """
    return bool value suggest whether the data can be read successfully
    return True is the data has problem
    return False if the data is ok
    """
    if path[-14:] == "/geon-info.txt":
        path = path[:-14]
    return cc.check_compressed_eigen_vectors(path)

@q.timer
def eigen_system_repartition(Coordinate new_size_node, str path, str path_new=""):
    if path[-14:] == "/geon-info.txt":
        path = path[:-14]
    if path_new[-14:] == "/geon-info.txt":
        path_new = path_new[:-14]
    return cc.eigen_system_repartition(new_size_node.xx, path, path_new)

## --------------

def show_all_shuffled_fields_writer():
    return cc.show_all_shuffled_fields_writer()



import numpy as np

class q:
    from qlat_utils import (
            timer_verbose,
            get_num_node,
            get_id_node,
            Coordinate,
            get_chunk_list,
            displayln_info,
            displayln,
            )
    from .c import (
            get_size_node,
            get_coor_node,
            begin,
            end,
            )

default_size_node_list = list(map(q.Coordinate, [
    [ 1, 1, 1, 1, ],
    [ 1, 1, 1, 2, ],
    [ 1, 1, 2, 2, ],
    [ 1, 2, 2, 2, ],
    [ 1, 2, 2, 2, ],
    [ 2, 2, 2, 2, ],
    [ 2, 2, 2, 4, ],
    [ 2, 2, 4, 4, ],
    [ 2, 4, 4, 4, ],
    [ 4, 4, 4, 4, ],
    [ 4, 4, 4, 8, ],
    [ 4, 4, 8, 8, ],
    [ 4, 8, 8, 8, ],
    [ 8, 8, 8, 8, ],
    [ 8, 8, 8, 16, ],
    [ 1, 1, 1, 3, ],
    [ 1, 1, 2, 3, ],
    [ 1, 2, 2, 3, ],
    [ 2, 2, 2, 3, ],
    [ 2, 2, 2, 6, ],
    [ 2, 2, 4, 6, ],
    [ 2, 4, 4, 6, ],
    [ 4, 4, 4, 6, ],
    [ 4, 4, 4, 12, ],
    [ 4, 4, 8, 12, ],
    [ 4, 8, 8, 12, ],
    [ 8, 8, 8, 12, ],
    ]))

comm = None

def set_comm(x):
    global comm
    comm = x

def get_comm():
    return comm

def begin_with_mpi(size_node_list=None):
    global comm
    from mpi4py import MPI
    comm = MPI.COMM_WORLD
    num_node = comm.size
    id_node = comm.rank
    if size_node_list is None:
        size_node_list = []
    else:
        assert isinstance(size_node_list, list)
        size_node_list = list(map(q.Coordinate, size_node_list))
    size_node_list = size_node_list + default_size_node_list
    size_node = None
    for size_node_check in size_node_list:
        if size_node_check.volume() == num_node:
            size_node = size_node_check
            break
    if size_node is None:
        if id_node == 0:
            q.displayln(size_node_list)
        comm.barrier()
        raise Exception("begin_with_mpi: size_node_list not match num_node")
    q.begin(id_node, size_node)

def end_with_mpi(is_preserving_cache=False):
    q.end(is_preserving_cache)
    from mpi4py import MPI
    MPI.Finalize()

@q.timer_verbose
def show_machine():
    q.displayln(f"id_node: {q.get_id_node():4} / {q.get_num_node()}"
            f" ; coor_node: {str(q.get_coor_node()):9}"
            f" / {str(q.get_size_node())}")

def get_mpi_chunk(total_list, *, rng_state=None):
    """
    rng_state has to be the same on all the nodes
    e.g. rng_state = q.RngState("get_mpi_chunk")
    """
    chunk_number = q.get_num_node()
    chunk_id = q.get_id_node()
    chunk_list = q.get_chunk_list(total_list, chunk_number=chunk_number, rng_state=rng_state)
    if chunk_id < len(chunk_list):
        return chunk_list[chunk_id]
    else:
        return []



# cython: binding=True, embedsignature=True, c_string_type=unicode, c_string_encoding=utf8

from qlat_utils.all cimport *
from . cimport everything as cc
from cpython cimport Py_buffer
from cpython.buffer cimport PyBUF_FORMAT

from .geometry cimport *
from .field_types cimport *
from .field_base cimport (
        SelectedPointsBase,
        SelectedFieldBase,
        FieldBase,
        )
from .selected_points_types cimport (
        SelectedPointsChar,
        )

import cqlat as c
import numpy as np

class q:
    from qlat_utils import (
            timer,
            get_id_node,
            get_num_node,
            hash_sha256,
            mk_cache,
            get_fname,
            displayln_info,
            )
    from .mpi_utils import (
            get_comm,
            )

### -------------------------------------------------------------------

cdef class SelectedShufflePlan:

    def __init__(self, *args):
        """
        SelectedShufflePlan()
        SelectedShufflePlan("l_from_g", psel_src, root)
        SelectedShufflePlan("l_from_g", psel_src_list, root_list)
        SelectedShufflePlan("g_from_l", psel_src, root, geo)
        SelectedShufflePlan("g_from_l", psel_src_list, root_list, geo_src_list)
        SelectedShufflePlan("r_from_l", psel_src, geo, rs)
        SelectedShufflePlan("r_from_l", psel_src_list, geo_src_list, rs)
        SelectedShufflePlan("dist_r_from_l", psel_src, geo, rs, id_node_list)
        SelectedShufflePlan("t_slice_from_l", psel_src_list, geo_src_list)
        SelectedShufflePlan("dist_t_slice_from_l", psel_src, geo, num_field)
        """
        self.psel_src_list = None
        self.psel_dst_list = None
        self.fsel_src_list = None
        self.fsel_dst_list = None
        self.geo_src_list = None
        self.geo_dst_list = None
        self.xx.init()
        if len(args) == 0:
            return
        elif len(args) == 1:
            return
        elif args[0] == "l_from_g" and isinstance(args[1], PointsSelection):
            self.init_from_psel_l_from_g(*args[1:])
        elif args[0] == "l_from_g" and isinstance(args[1], list) and len(args[1]) > 0 and isinstance(args[1][0], PointsSelection):
            self.init_from_psel_list_l_from_g(*args[1:])
        elif args[0] == "g_from_l" and isinstance(args[1], PointsSelection):
            self.init_from_psel_g_from_l(*args[1:])
        elif args[0] == "g_from_l" and isinstance(args[1], list) and len(args[1]) > 0 and isinstance(args[1][0], PointsSelection):
            self.init_from_psel_list_g_from_l(*args[1:])
        elif args[0] == "r_from_l" and isinstance(args[1], PointsSelection):
            self.init_from_psel_r_from_l(*args[1:])
        elif args[0] == "r_from_l" and isinstance(args[1], list) and len(args[1]) > 0 and isinstance(args[1][0], PointsSelection):
            self.init_from_psel_list_r_from_l(*args[1:])
        elif args[0] == "dist_r_from_l" and isinstance(args[1], PointsSelection):
            self.init_from_psel_dist_r_from_l(*args[1:])
        elif args[0] == "t_slice_from_l" and isinstance(args[1], list) and len(args[1]) > 0 and isinstance(args[1][0], PointsSelection):
            self.init_from_psel_list_t_slice_from_l(*args[1:])
        elif args[0] == "dist_t_slice_from_l" and isinstance(args[1], PointsSelection):
            self.init_from_psel_list_dist_t_slice_from_l(*args[1:])
        else:
            raise Exception(f"SelectedShufflePlan.__init__ {args}")

    @q.timer
    def init_from_psel_l_from_g(self, PointsSelection psel_src, cc.Int root):
        """
        shuffle to PointsDistType::Local ("l") from PointsDistType::Global ("g").
        """
        cdef cc.Int id_node = cc.get_id_node()
        cdef PointsSelection psel_dst = PointsSelection()
        self.xx.init()
        cc.set_selected_shuffle_plan_l_from_g(self.xx, psel_src.xx, root)
        psel_dst.xx.points_dist_type = self.xx.points_dist_type_recv
        cc.shuffle_points_selection(psel_dst.xx, psel_src.xx, self.xx)
        if root == id_node:
            self.psel_src_list = [ psel_src, ]
        else:
            self.psel_src_list = [ None, ]
        self.psel_dst_list = [ psel_dst, ]

    @q.timer
    def init_from_psel_list_l_from_g(self, list psel_src_list, list root_list):
        """
        shuffle to PointsDistType::Local ("l") from PointsDistType::Global ("g").
        """
        assert len(root_list) == len(psel_src_list)
        cdef cc.Int id_node = cc.get_id_node()
        cdef cc.Int root
        cdef cc.std_vector[cc.Int] root_vec
        root_vec.resize(len(root_list))
        for i in range(root_vec.size()):
            root = root_list[i]
            root_vec[i] = root
        self.psel_src_list = []
        for i in range(root_vec.size()):
            root = root_vec[i]
            if root == id_node:
                self.psel_src_list.append(psel_src_list[i])
            else:
                self.psel_src_list.append(None)
        cdef cc.std_vector[cc.PointsSelection] psel_src_vec
        psel_src_vec.resize(len(self.psel_src_list))
        cdef PointsSelection psel
        for i in range(psel_src_vec.size()):
            psel = self.psel_src_list[i]
            if psel is not None:
                cc.qswap(psel.xx, psel_src_vec[i])
        self.xx.init()
        cc.set_selected_shuffle_plan_l_from_g(self.xx, psel_src_vec, root_vec)
        cdef cc.std_vector[cc.PointsSelection] psel_dst_vec
        cc.shuffle_points_selection(psel_dst_vec, psel_src_vec, self.xx)
        psel_dst_list = [ PointsSelection() for i in range(psel_dst_vec.size()) ]
        for i in range(psel_src_vec.size()):
            psel = self.psel_src_list[i]
            if psel is not None:
                cc.qswap(psel.xx, psel_src_vec[i])
        for i in range(psel_dst_vec.size()):
            psel = psel_dst_list[i]
            cc.qswap(psel.xx, psel_dst_vec[i])
        self.psel_dst_list = psel_dst_list

    @q.timer
    def init_from_psel_g_from_l(self, PointsSelection psel_src, cc.Int root, Geometry geo_src):
        """
        shuffle to PointsDistType::Global ("g") from PointsDistType::Local ("l").
        """
        cdef PointsSelection psel_dst = PointsSelection()
        self.xx.init()
        cc.set_selected_shuffle_plan_g_from_l(self.xx, psel_src.xx, root, geo_src.xx)
        psel_dst.xx.points_dist_type = self.xx.points_dist_type_recv
        cc.shuffle_points_selection(psel_dst.xx, psel_src.xx, self.xx)
        self.psel_src_list = [ psel_src, ]
        self.psel_dst_list = [ psel_dst, ]

    @q.timer
    def init_from_psel_list_g_from_l(self, list psel_src_list, list root_list, list geo_src_list):
        """
        shuffle to PointsDistType::Global ("g") from PointsDistType::Local ("l").
        """
        assert len(root_list) == len(psel_src_list)
        assert len(geo_src_list) == len(psel_src_list)
        cdef cc.std_vector[cc.PointsSelection] psel_src_vec
        psel_src_vec.resize(len(psel_src_list))
        cdef PointsSelection psel
        for i in range(psel_src_vec.size()):
            psel = psel_src_list[i]
            cc.qswap(psel.xx, psel_src_vec[i])
        cdef cc.Int root
        cdef cc.std_vector[cc.Int] root_vec
        root_vec.resize(len(root_list))
        for i in range(root_vec.size()):
            root = root_list[i]
            root_vec[i] = root
        cdef cc.std_vector[cc.Geometry] geo_src_vec
        geo_src_vec.resize(len(geo_src_list))
        cdef Geometry geo
        for i in range(geo_src_vec.size()):
            geo = geo_src_list[i]
            geo_src_vec[i] = geo.xx
        self.xx.init()
        cc.set_selected_shuffle_plan_g_from_l(self.xx, psel_src_vec, root_vec, geo_src_vec)
        cdef cc.std_vector[cc.PointsSelection] psel_dst_vec
        cc.shuffle_points_selection(psel_dst_vec, psel_src_vec, self.xx)
        psel_dst_list = [ PointsSelection() for i in range(psel_dst_vec.size()) ]
        for i in range(psel_src_vec.size()):
            psel = psel_src_list[i]
            cc.qswap(psel.xx, psel_src_vec[i])
        for i in range(psel_dst_vec.size()):
            psel = psel_dst_list[i]
            cc.qswap(psel.xx, psel_dst_vec[i])
        self.psel_src_list = psel_src_list
        self.psel_dst_list = psel_dst_list

    @q.timer
    def init_from_psel_r_from_l(self, PointsSelection psel_src, Geometry geo_src, RngState rs):
        """
        shuffle to PointsDistType::Random ("r") from PointsDistType::Local ("l").
        """
        cdef PointsSelection psel_dst = PointsSelection()
        self.xx.init()
        cc.set_selected_shuffle_plan_r_from_l(self.xx, psel_src.xx, geo_src.xx, rs.xx)
        psel_dst.xx.points_dist_type = self.xx.points_dist_type_recv
        cc.shuffle_points_selection(psel_dst.xx, psel_src.xx, self.xx)
        self.psel_src_list = [ psel_src, ]
        self.psel_dst_list = [ psel_dst, ]

    @q.timer
    def init_from_psel_list_r_from_l(self, list psel_src_list, list geo_src_list, RngState rs):
        """
        shuffle to PointsDistType::Random ("r") from PointsDistType::Local ("l").
        """
        assert len(geo_src_list) == len(psel_src_list)
        cdef cc.std_vector[cc.PointsSelection] psel_src_vec
        psel_src_vec.resize(len(psel_src_list))
        cdef PointsSelection psel
        for i in range(psel_src_vec.size()):
            psel = psel_src_list[i]
            cc.qswap(psel.xx, psel_src_vec[i])
        cdef cc.std_vector[cc.Geometry] geo_src_vec
        geo_src_vec.resize(len(geo_src_list))
        cdef Geometry geo
        for i in range(geo_src_vec.size()):
            geo = geo_src_list[i]
            geo_src_vec[i] = geo.xx
        self.xx.init()
        cc.set_selected_shuffle_plan_r_from_l(self.xx, psel_src_vec, geo_src_vec, rs.xx)
        cdef cc.std_vector[cc.PointsSelection] psel_dst_vec
        cc.shuffle_points_selection(psel_dst_vec, psel_src_vec, self.xx)
        psel_dst_list = [ PointsSelection() for i in range(psel_dst_vec.size()) ]
        for i in range(psel_src_vec.size()):
            psel = psel_src_list[i]
            cc.qswap(psel.xx, psel_src_vec[i])
        for i in range(psel_dst_vec.size()):
            psel = psel_dst_list[i]
            cc.qswap(psel.xx, psel_dst_vec[i])
        self.psel_src_list = psel_src_list
        self.psel_dst_list = psel_dst_list

    @q.timer
    def init_from_psel_dist_r_from_l(self, PointsSelection psel_src, Geometry geo_src, RngState rs, list id_node_list=None):
        """
        shuffle to PointsDistType::Random ("r") (limited to ``id_node_list``) from PointsDistType::Local ("l").
        """
        cdef cc.std_vector[cc.PointsSelection] psel_src_vec
        psel_src_vec.resize(1)
        cc.qswap(psel_src.xx, psel_src_vec[0])
        cdef cc.Int id_node
        cdef cc.std_vector[cc.Int] id_node_vec
        if id_node_list is not None:
            id_node_vec.resize(len(id_node_list))
            for i in range(id_node_vec.size()):
                id_node = id_node_list[i]
                id_node_vec[i] = id_node
        self.xx.init()
        cc.set_selected_shuffle_plan_dist_r_from_l(self.xx, psel_src.xx, geo_src.xx, rs.xx, id_node_vec)
        cdef cc.std_vector[cc.PointsSelection] psel_dst_vec
        cc.shuffle_points_selection(psel_dst_vec, psel_src_vec, self.xx)
        psel_dst_list = [ PointsSelection() for i in range(psel_dst_vec.size()) ]
        cc.qswap(psel_src.xx, psel_src_vec[0])
        cdef PointsSelection psel
        for i in range(psel_dst_vec.size()):
            psel = psel_dst_list[i]
            cc.qswap(psel.xx, psel_dst_vec[i])
        self.psel_src_list = [ psel_src, ]
        self.psel_dst_list = psel_dst_list

    @q.timer
    def init_from_psel_list_t_slice_from_l(self, list psel_src_list, list geo_src_list):
        """
        shuffle to PointsDistType::Local ("l") from PointsDistType::Local ("l").
        """
        assert len(geo_src_list) == len(psel_src_list)
        cdef cc.std_vector[cc.PointsSelection] psel_src_vec
        cdef cc.std_vector[cc.PointsSelection] psel_dst_vec
        cdef PointsSelection psel
        psel_src_vec.resize(len(psel_src_list))
        for i in range(psel_src_vec.size()):
            psel = psel_src_list[i]
            cc.qswap(psel.xx, psel_src_vec[i])
        cdef cc.std_vector[cc.Geometry] geo_src_vec
        geo_src_vec.resize(len(geo_src_list))
        cdef Geometry geo
        for i in range(geo_src_vec.size()):
            geo = geo_src_list[i]
            geo_src_vec[i] = geo.xx
        self.xx.init()
        cc.set_selected_shuffle_plan_t_slice_from_l(self.xx, psel_src_vec, geo_src_vec)
        cc.shuffle_points_selection(psel_dst_vec, psel_src_vec, self.xx)
        psel_dst_list = [ PointsSelection() for i in range(psel_dst_vec.size()) ]
        for i in range(psel_src_vec.size()):
            psel = psel_src_list[i]
            cc.qswap(psel.xx, psel_src_vec[i])
        for i in range(psel_dst_vec.size()):
            psel = psel_dst_list[i]
            cc.qswap(psel.xx, psel_dst_vec[i])
        self.psel_src_list = psel_src_list
        self.psel_dst_list = psel_dst_list

    @q.timer
    def init_from_psel_list_dist_t_slice_from_l(self, PointsSelection psel_src, Geometry geo, int num_field):
        """
        shuffle to PointsDistType::Local ("l") from PointsDistType::Local ("l").
        match ``init_from_psel_list_t_slice_from_l``.
        Example use case:
            Perform spatial smearing for propagators (see ``q.prop_spatial_smear``):
            (1) shuffle propagators with "t_slice_from_l".
            (2) shuffle gauge field with "dist_t_slice_from_l".
        """
        cdef cc.std_vector[cc.PointsSelection] psel_src_vec
        cdef cc.std_vector[cc.PointsSelection] psel_dst_vec
        cdef PointsSelection psel
        psel_src_vec.resize(1)
        cc.qswap(psel_src.xx, psel_src_vec[0])
        self.xx.init()
        cc.set_selected_shuffle_plan_dist_t_slice_from_l(self.xx, psel_src_vec[0], geo.xx, num_field)
        cc.shuffle_points_selection(psel_dst_vec, psel_src_vec, self.xx)
        psel_dst_list = [ PointsSelection() for i in range(psel_dst_vec.size()) ]
        cc.qswap(psel_src.xx, psel_src_vec[0])
        for i in range(psel_dst_vec.size()):
            psel = psel_dst_list[i]
            cc.qswap(psel.xx, psel_dst_vec[i])
        self.psel_src_list = [ psel_src, ]
        self.psel_dst_list = psel_dst_list

    @property
    def points_dist_type_send(self):
        return cc.show(self.xx.points_dist_type_send)

    @property
    def points_dist_type_recv(self):
        return cc.show(self.xx.points_dist_type_recv)

    @property
    def num_selected_points_send(self):
        return self.xx.num_selected_points_send

    @property
    def num_selected_points_recv(self):
        return self.xx.num_selected_points_recv

    @property
    def psel_send_list(self):
        return self.psel_src_list

    @property
    def psel_recv_list(self):
        return self.psel_dst_list

    @property
    def geo_send_list(self):
        if self.geo_src_list is not None:
            return self.geo_src_list
        self.geo_src_list = []
        assert self.xx.size_node_send.size() == self.xx.coor_node_send.size()
        cdef cc.Int num_geo = self.xx.size_node_send.size()
        cdef Coordinate total_site = Coordinate()
        total_site.xx = self.xx.total_site
        cdef Coordinate size_node = Coordinate()
        cdef Coordinate coor_node = Coordinate()
        cdef Coordinate node_site = Coordinate()
        cdef Geometry geo
        if total_site == Coordinate():
            return self.geo_src_list
        cdef list geo_list = []
        for i in range(num_geo):
            size_node.xx = self.xx.size_node_send[i]
            coor_node.xx = self.xx.coor_node_send[i]
            node_site.xx = total_site.xx / size_node.xx
            assert node_site * size_node == total_site
            geo = Geometry(coor_node, size_node, node_site)
            geo_list.append(geo)
        self.geo_src_list = geo_list
        return self.geo_src_list

    @property
    def geo_recv_list(self):
        if self.geo_dst_list is not None:
            return self.geo_dst_list
        self.geo_dst_list = []
        assert self.xx.size_node_recv.size() == self.xx.coor_node_recv.size()
        cdef cc.Int num_geo = self.xx.size_node_recv.size()
        cdef Coordinate total_site = Coordinate()
        total_site.xx = self.xx.total_site
        cdef Coordinate size_node = Coordinate()
        cdef Coordinate coor_node = Coordinate()
        cdef Coordinate node_site = Coordinate()
        cdef Geometry geo
        if total_site == Coordinate():
            return self.geo_dst_list
        cdef list geo_list = []
        for i in range(num_geo):
            size_node.xx = self.xx.size_node_recv[i]
            coor_node.xx = self.xx.coor_node_recv[i]
            node_site.xx = total_site.xx / size_node.xx
            assert node_site * size_node == total_site
            geo = Geometry(coor_node, size_node, node_site)
            geo_list.append(geo)
        self.geo_dst_list = geo_list
        return self.geo_dst_list

    @property
    def fsel_send_list(self):
        if self.fsel_src_list is not None:
            return self.fsel_src_list
        self.fsel_src_list = []
        if self.points_dist_type_send not in [ "l", "f", ]:
            return self.fsel_src_list
        if len(self.psel_send_list) != len(self.geo_send_list):
            return self.fsel_src_list
        cdef FieldSelection fsel
        cdef list fsel_list = []
        for psel, geo in zip(self.psel_send_list, self.geo_send_list):
            fsel = FieldSelection(psel, geo)
            fsel_list.append(fsel)
        self.fsel_src_list = fsel_list
        return self.fsel_src_list

    @property
    def fsel_recv_list(self):
        if self.fsel_dst_list is not None:
            return self.fsel_dst_list
        self.fsel_dst_list = []
        if self.points_dist_type_recv not in [ "l", "f", ]:
            return self.fsel_dst_list
        if len(self.psel_recv_list) != len(self.geo_recv_list):
            return self.fsel_dst_list
        cdef FieldSelection fsel
        cdef list fsel_list = []
        for psel, geo in zip(self.psel_recv_list, self.geo_recv_list):
            fsel = FieldSelection(psel, geo)
            fsel_list.append(fsel)
        self.fsel_dst_list = fsel_list
        return self.fsel_dst_list

    @q.timer
    def shuffle(self, SelectedPointsChar sp_src, *, bint is_reverse=False):
        """
        shuffle `sp_src` with this plan
        return `sp_dst`
        The type for all `sp` is `SelectedPointsChar`
        """
        assert 1 == len(self.psel_send_list)
        assert 1 == len(self.psel_recv_list)
        assert isinstance(sp_src, SelectedPointsChar)
        psel_src = self.psel_send_list[0]
        psel_dst = self.psel_recv_list[0]
        psel = sp_src.psel
        cdef SelectedPointsChar sp_dst
        if is_reverse:
            assert (psel is None) or (psel_dst is None) or (psel == psel_dst)
            sp_dst = SelectedPointsChar(psel_src)
            cc.shuffle_selected_points_back(sp_dst.xx, sp_src.xx, self.xx)
        else:
            assert (psel is None) or (psel_src is None) or(psel == psel_src)
            sp_dst = SelectedPointsChar(psel_dst)
            cc.shuffle_selected_points(sp_dst.xx, sp_src.xx, self.xx)
        return sp_dst

    @q.timer
    def shuffle_list(self, list sp_src_list, *, bint is_reverse=False):
        """
        shuffle `sp_src_list` with this plan
        return `sp_dst_list`
        The type for all `sp` is `SelectedPointsChar`
        """
        if is_reverse:
            psel_src_list = self.psel_recv_list
            psel_dst_list = self.psel_send_list
        else:
            psel_src_list = self.psel_send_list
            psel_dst_list = self.psel_recv_list
        assert len(sp_src_list) == len(psel_src_list)
        for sp_src, psel_src in zip(sp_src_list, psel_src_list):
            assert isinstance(sp_src, SelectedPointsChar)
            psel = sp_src.psel
            assert (psel is None) or (psel_src is None) or (psel == psel_src)
        cdef cc.std_vector[cc.SelectedPoints[cc.Char]] sp_src_vec
        cdef cc.std_vector[cc.SelectedPoints[cc.Char]] sp_dst_vec
        cdef SelectedPointsChar sp
        sp_src_vec.resize(len(psel_src_list))
        for i in range(sp_src_vec.size()):
            sp = sp_src_list[i]
            cc.qswap(sp.xx, sp_src_vec[i])
        if is_reverse:
            cc.shuffle_selected_points_back(sp_dst_vec, sp_src_vec, self.xx)
        else:
            cc.shuffle_selected_points(sp_dst_vec, sp_src_vec, self.xx)
        assert <cc.Long>sp_dst_vec.size() == len(psel_dst_list)
        sp_dst_list = [ SelectedPointsChar(psel) for psel in psel_dst_list ]
        for i in range(sp_src_vec.size()):
            sp = sp_src_list[i]
            cc.qswap(sp.xx, sp_src_vec[i])
        for i in range(sp_dst_vec.size()):
            sp = sp_dst_list[i]
            assert sp.psel is psel_dst_list[i]
            cc.qswap(sp.xx, sp_dst_vec[i])
        return sp_dst_list

    @q.timer
    def shuffle_sp(self, cls, object src, *, bint is_reverse=False):
        """
        shuffle `src` with this plan
        return `sp_dst`
        The type for the returned object is `cls`
        """
        assert len(self.psel_send_list) == 1
        assert len(self.psel_recv_list) == 1
        cdef SelectedPointsChar spc_src
        cdef SelectedPointsChar spc_dst
        cdef Geometry geo
        cdef list geo_send_list
        cdef list geo_recv_list
        if is_reverse:
            geo_send_list = self.geo_recv_list
            geo_recv_list = self.geo_send_list
        else:
            geo_send_list = self.geo_send_list
            geo_recv_list = self.geo_recv_list
        cdef Geometry geo_send = None
        cdef Geometry geo_recv = None
        if geo_send_list is not None:
            if len(geo_send_list) == 1:
                geo_send = geo_send_list[0]
        if geo_recv_list is not None:
            if len(geo_recv_list) == 1:
                geo_recv = geo_recv_list[0]
        spc_src = SelectedPointsChar()
        if isinstance(src, SelectedPointsBase):
            src.swap_cast(spc_src)
            spc_dst = self.shuffle(spc_src, is_reverse=is_reverse)
            src.swap_cast(spc_src)
        elif isinstance(src, SelectedFieldBase):
            geo = Geometry()
            src.swap_sp_cast(spc_src, geo)
            if geo_send is not None:
                assert geo == geo_send
            spc_dst = self.shuffle(spc_src, is_reverse=is_reverse)
            src.swap_sp_cast(spc_src, geo)
        elif isinstance(src, FieldBase):
            geo = Geometry()
            src.swap_sp_cast(spc_src, geo)
            if geo_send is not None:
                assert geo == geo_send
            spc_dst = self.shuffle(spc_src, is_reverse=is_reverse)
            src.swap_sp_cast(spc_src, geo)
        else:
            assert False
        dst = cls()
        if geo_recv is None:
            geo_recv = Geometry()
        geo = geo_recv.copy()
        if isinstance(src, SelectedPointsBase):
            dst.swap_cast(spc_dst)
            if is_reverse:
                psel_list = self.psel_send_list
            else:
                psel_list = self.psel_recv_list
            assert len(psel_list) == 1
            dst.psel = psel_list[0]
        elif isinstance(src, SelectedFieldBase):
            dst.swap_sp_cast(spc_dst, geo)
            if is_reverse:
                fsel_list = self.fsel_send_list
            else:
                fsel_list = self.fsel_recv_list
            if len(fsel_list) == 1:
                dst.fsel = fsel_list[0]
        elif isinstance(src, FieldBase):
            dst.swap_sp_cast(spc_dst, geo)
        else:
            assert False
        return dst

    @q.timer
    def shuffle_sp_list(self, cls, list src_list, *, bint is_reverse=False):
        """
        shuffle `src_list` with this plan
        return `sp_dst_list`
        The type for the returned list of objects are `cls`
        """
        cdef list geo_send_list
        cdef list geo_recv_list
        cdef list psel_send_list
        cdef list psel_recv_list
        cdef int num_send
        cdef int num_recv
        if is_reverse:
            geo_send_list = self.geo_recv_list
            geo_recv_list = self.geo_send_list
            psel_send_list = self.psel_recv_list
            psel_recv_list = self.psel_send_list
            num_send = self.num_selected_points_recv
            num_recv = self.num_selected_points_send
        else:
            geo_send_list = self.geo_send_list
            geo_recv_list = self.geo_recv_list
            psel_send_list = self.psel_send_list
            psel_recv_list = self.psel_recv_list
            num_send = self.num_selected_points_send
            num_recv = self.num_selected_points_recv
        assert len(psel_send_list) == num_send
        assert len(psel_recv_list) == num_recv
        assert len(src_list) == num_send
        if len(geo_send_list) != 0:
            assert len(geo_send_list) == num_send
        else:
            geo_send_list = [ None for i in range(num_send) ]
        if len(geo_recv_list) != 0:
            assert len(geo_recv_list) == num_recv
        else:
            geo_recv_list = [ None for i in range(num_recv) ]
        spc_src_list = []
        geo_src_list = []
        for idx, (src, geo_send,) in enumerate(zip(src_list, geo_send_list)):
            spc_src = SelectedPointsChar()
            geo_src = Geometry()
            if isinstance(src, SelectedPointsBase):
                src.swap_cast(spc_src)
                if src.psel is not None:
                    assert src.psel == psel_send_list[idx]
            elif isinstance(src, SelectedFieldBase):
                src.swap_sp_cast(spc_src, geo_src)
                assert geo_send is not None
                assert geo_src == geo_send
                if is_reverse:
                    fsel_list = self.fsel_recv_list
                else:
                    fsel_list = self.fsel_send_list
                assert len(fsel_list) == num_send
                if src.fsel is not None:
                    if src.fsel is not fsel_list[idx]:
                        assert PointsSelection(src.fsel) == self.psel_send_list[idx]
            elif isinstance(src, FieldBase):
                src.swap_sp_cast(spc_src, geo_src)
                assert geo_send is not None
                assert geo_src == geo_send
                assert spc_src.n_points == geo_send.local_volume_expanded
            else:
                assert False
            spc_src_list.append(spc_src)
            geo_src_list.append(geo_src)
        assert len(spc_src_list) == num_send
        spc_dst_list = self.shuffle_list(spc_src_list, is_reverse=is_reverse)
        assert len(spc_dst_list) == num_recv
        for src, spc_src, geo_src in zip(src_list, spc_src_list, geo_src_list):
            if isinstance(src, SelectedPointsBase):
                src.swap_cast(spc_src)
            elif isinstance(src, SelectedFieldBase):
                src.swap_sp_cast(spc_src, geo_src)
            elif isinstance(src, FieldBase):
                src.swap_sp_cast(spc_src, geo_src)
            else:
                assert False
        dst_list = []
        cdef Geometry geo
        for idx, (spc_dst, geo_recv,) in enumerate(zip(spc_dst_list, geo_recv_list)):
            dst = cls()
            if isinstance(src, SelectedPointsBase):
                dst.swap_cast(spc_dst)
                assert dst.psel is psel_recv_list[idx]
            elif isinstance(src, SelectedFieldBase):
                assert geo_recv is not None
                geo = geo_recv.copy()
                dst.swap_sp_cast(spc_dst, geo)
                if is_reverse:
                    fsel_list = self.fsel_send_list
                else:
                    fsel_list = self.fsel_recv_list
                assert len(fsel_list) == num_recv
                dst.fsel = fsel_list[idx]
            elif isinstance(src, FieldBase):
                assert geo_recv is not None
                geo = geo_recv.copy()
                assert geo.local_volume_expanded == spc_dst.n_points
                dst.swap_sp_cast(spc_dst, geo)
            else:
                assert False
            dst_list.append(dst)
        assert len(dst_list) == num_recv
        return dst_list

###

cdef class PointsSelection:

    def __cinit__(self):
        self.cdata = <cc.Long>&(self.xx)
        self.view_count = 0

    def __init__(self, *args):
        """
        PointsSelection()
        PointsSelection(total_site)
        PointsSelection(total_site, xg_arr)
        PointsSelection(total_site, xg_list)
        PointsSelection(total_site, xg)
        PointsSelection(total_site, n_points)
        PointsSelection(total_site, xg_arr, points_dist_type)
        PointsSelection(geo) # full selection (type "f")
        PointsSelection(fsel) # (type "l")
        PointsSelection(fsel, ssp)
        PointsSelection(psel) # A copy of `psel`.
        PointsSelection(psel, ssp)
        PointsSelection(psel, ssp, is_reverse)
        #
        points_dist_type in [ "g", "f", "l", "r", "o", ]
        """
        cdef cc.Int len_args = len(args)
        if len_args == 0:
            self.xx.init()
        elif isinstance(args[0], Coordinate):
            total_site = args[0]
            self.init_from_total_site(*args)
        elif isinstance(args[0], Geometry):
            self.init_from_geo(*args)
        elif isinstance(args[0], FieldSelection):
            self.init_from_fsel(*args)
        elif isinstance(args[0], PointsSelection):
            self.init_from_psel(*args)
        else:
            raise Exception(f"PointsSelection::__init__: {args}")

    def init_from_total_site(self, Coordinate total_site, object xg_arr=None, str points_dist_type=None):
        """
        ``xg_arr`` can be n_points, xg, xg_arr, xg_list.
        default ``points_dist_type == "g"``
        ``points_dist_type in [ "g", "f", "l", "r", "o", ]``
        """
        self.xx.init()
        self.xx.init(total_site.xx, 0)
        if xg_arr is not None:
            self.xg_arr = xg_arr
        if points_dist_type is not None:
            self.points_dist_type = points_dist_type

    def init_from_geo(self, Geometry geo):
        """
        points_dist_type in [ "f", ]
        """
        cc.set_psel_full(self.xx, geo.xx)

    def init_from_psel(self, PointsSelection psel, SelectedShufflePlan ssp=None, bint is_reverse=False):
        """
        Shuffle according to `ssp`.
        """
        fname = q.get_fname()
        self.xx.init()
        if ssp is None:
            self @= psel
        elif (len(ssp.psel_send_list) == 1 and (ssp.psel_send_list[0] is psel or ssp.psel_send_list[0] is None)) and (not is_reverse):
            assert len(ssp.psel_recv_list) == 1
            self @= ssp.psel_recv_list[0]
        elif (len(ssp.psel_recv_list) == 1 and ssp.psel_recv_list[0] is psel) and is_reverse:
            assert len(ssp.psel_send_list) == 1
            if ssp.psel_send_list[0] is not None:
                self @= ssp.psel_send_list[0]
        else:
            q.displayln_info(f"WARNING: {fname}: psel is not ssp.psel_src")
            if is_reverse:
                assert cc.show(psel.xx.points_dist_type) == cc.show(ssp.xx.points_dist_type_recv)
                self.xx.points_dist_type = ssp.xx.points_dist_type_send
                cc.shuffle_points_selection_back(self.xx, psel.xx, ssp.xx)
            else:
                assert cc.show(psel.xx.points_dist_type) == cc.show(ssp.xx.points_dist_type_send)
                self.xx.points_dist_type = ssp.xx.points_dist_type_recv
                cc.shuffle_points_selection(self.xx, psel.xx, ssp.xx)

    def init_from_fsel(self, FieldSelection fsel, SelectedShufflePlan ssp=None):
        """
        Shuffle according to `ssp`.
        self.points_dist_type == "l" for PointsDistType::Local (if ssp is None)
        """
        self.xx.init()
        if ssp is None:
            cc.set_psel_from_fsel(self.xx, fsel.xx)
        else:
            self.xx.points_dist_type = ssp.xx.points_dist_type_recv
            cc.shuffle_field_selection(self.xx, fsel.xx, ssp.xx)

    def __getbuffer__(self, Py_buffer *buffer, int flags):
        cdef int ndim = 2
        cdef Buffer buf = Buffer(self, ndim)
        buf.format = 'i'
        buf.itemsize = sizeof(cc.Int)
        buf.buf = <char*>(self.xx.data())
        buf.set_dim_size(0, self.xx.size())
        buf.set_dim_size(1, 4)
        buf.update_strides_from_shape()
        buf.set_buffer(buffer, flags)
        self.view_count += 1

    def release_buffer(self, Buffer buf):
        assert buf.obj is self
        self.view_count -= 1

    def __imatmul__(self, PointsSelection v1 not None):
        if self.view_count > 0:
            raise Exception("PointsSelection.__imatmul__: self.view_count>0")
        cc.assign_direct(self.xx, v1.xx)
        return self

    def copy(self):
        x = type(self)()
        x @= self
        return x

    def __copy__(self):
        return self.copy()

    def __deepcopy__(self, memo):
        return self.copy()

    def swap(self, PointsSelection other):
        cc.qswap(self.xx, other.xx)

    def swap_sp_cast(self, SelectedPointsChar other, Coordinate total_site):
        assert other.psel is None
        cc.qswap_cast(self.xx, other.xx, total_site.xx)

    @property
    def points_dist_type(self):
        """
        points_dist_type in [ "g", "f", "l", "r", "o", ]
        Meaning: Global, Full, Local, Random, Other
        """
        return cc.show(self.xx.points_dist_type)

    @points_dist_type.setter
    def points_dist_type(self, str value):
        """
        set the points_dist_type flag
        value in [ "g", "f", "l", "r", "o", ]
        """
        self.xx.points_dist_type = cc.read_points_dist_type(value)

    @property
    def total_site(self):
        cdef Coordinate value = Coordinate()
        value.xx = self.xx.total_site
        return value

    @total_site.setter
    def total_site(self, Coordinate value):
        self.xx.total_site = value.xx

    @property
    def geo(self):
        cdef Geometry geo = Geometry(self.total_site)
        return geo

    @property
    def n_points(self):
        return self.xx.size()

    @property
    def xg_arr(self):
        """
        return xg for all selected points
        shape = (psel.n_points, 4,)
        """
        return np.asarray(self, dtype=np.int32)

    @xg_arr.setter
    def xg_arr(self, object xg_arr):
        """
        xg_arr can be n_points, xg, xg_arr, xg_list.
        """
        if self.view_count > 0:
            raise ValueError("can't re-init while being viewed")
        cdef cc.Coordinate total_site = self.xx.total_site
        cdef cc.PointsDistType points_dist_type = self.xx.points_dist_type
        cdef cc.Long n_points
        cdef cc.Long i
        cdef Coordinate xg
        if isinstance(xg_arr, int):
            n_points = xg_arr
            self.xx.init(total_site, n_points, points_dist_type)
        elif isinstance(xg_arr, Coordinate):
            xg = xg_arr
            n_points = 1
            self.xx.init(total_site, n_points, points_dist_type)
            self.xx[0] = xg.xx
        elif isinstance(xg_arr, np.ndarray):
            n_points = len(xg_arr)
            assert xg_arr.shape == (n_points, 4,)
            self.xx.init(total_site, n_points, points_dist_type)
            np.asarray(self, dtype=np.int32)[:] = xg_arr
        elif isinstance(xg_arr, list):
            n_points = len(xg_arr)
            self.xx.init(total_site, n_points, points_dist_type)
            for i in range(n_points):
                self.xx[i] = Coordinate(xg_arr[i]).xx
        else:
            raise Exception(f"PointsSelection.xg_arr = {xg_arr}")

    def set_rand(self, Coordinate total_site not None, cc.Long n_points, RngState rs not None):
        if self.view_count > 0:
            raise ValueError("can't re-init while being viewed")
        cc.assign_direct(self.xx, cc.mk_random_points_selection(total_site.xx, n_points, rs.xx))

    def to_lat_data(self):
        cdef LatDataInt ld = LatDataInt()
        cc.lat_data_from_points_selection(ld.xx, self.xx)
        return ld

    def from_lat_data(self, LatDataInt ld not None):
        cc.points_selection_from_lat_data(self.xx, ld.xx)

    @q.timer
    def bcast_via_ld(self, cc.Int root=0):
        cdef LatDataInt ld
        if cc.get_num_node() != 1:
            if cc.get_id_node() == root:
                ld = self.to_lat_data()
            else:
                ld = LatDataInt()
            ld.bcast(root)
            if cc.get_id_node() != root:
                self.from_lat_data(ld)
        return self

    @q.timer
    def bcast(self, cc.Int root=0):
        cdef cc.Int ret = cc.bcast(self.xx, root)
        assert ret == 0
        return self

    def save_str(self):
        """
        only return str at node 0
        """
        cdef LatDataInt ld
        if q.get_id_node() == 0:
            ld = self.to_lat_data()
            return ld.save_str()
        else:
            return bytes()

    def load_str(self, bytes content):
        """
        only need str at node 0
        """
        cdef LatDataInt ld = LatDataInt()
        if q.get_id_node() == 0:
            ld.load_str(content)
        ld.bcast()
        self.from_lat_data(ld)

    def save(self, const cc.std_string& path, *, is_sync_node=True):
        if is_sync_node:
            cc.save_points_selection_info(self.xx, path)
        else:
            cc.save_points_selection(self.xx, path)

    def load(self, const cc.std_string& path, Geometry geo=None, *, is_sync_node=True):
        if self.view_count > 0:
            raise ValueError("can't re-init while being viewed")
        if is_sync_node:
            cc.assign_direct(self.xx, cc.load_points_selection_info(path))
        else:
            cc.assign_direct(self.xx, cc.load_points_selection(path))
        cdef Coordinate total_site
        if path.endswith(".lati"):
            if geo is not None:
                total_site = geo.total_site
                assert self.xx.total_site == total_site.xx
        else:
            assert self.xx.total_site == cc.Coordinate()
            assert geo is not None
            total_site = geo.total_site
            self.xx.total_site = total_site.xx

    def __setitem__(self, idx, val):
        """
        Implemented in terms of ``np.asarray``
        """
        np.asarray(self)[idx] = val

    def __getitem__(self, idx):
        """
        Implemented in terms of ``np.asarray``
        """
        return np.asarray(self)[idx]

    def __iter__(self):
        cdef cc.Long idx
        cdef cc.Long n_points = self.n_points
        for idx in range(n_points):
            yield self.coordinate_from_idx(idx)

    def __len__(self):
        return self.xx.size()

    def __eq__(self, PointsSelection other):
        if self is other:
            return True
        if self.xx == other.xx:
            return True
        return False

    def coordinate_from_idx(self, cc.Long idx):
        cdef Coordinate xg = Coordinate()
        xg.xx = self.xx[idx]
        return xg

    def intersect(self, FieldSelection fsel):
        """
        return new psel
        """
        cdef PointsSelection psel_new = self.copy()
        psel_new.xx = cc.intersect(fsel.xx, self.xx)
        return psel_new

    def is_containing_psel(self, PointsSelection psel_small):
        cdef cc.Bool x = cc.is_containing(self.xx, psel_small.xx)
        return x

    def is_containing_fsel(self, FieldSelection fsel_small):
        cdef cc.Bool x = cc.is_containing(self.xx, fsel_small.xx)
        return x

    def is_containing(self, sel_small):
        if isinstance(sel_small, PointsSelection):
            return self.is_containing_psel(sel_small)
        elif isinstance(sel_small, FieldSelection):
            return self.is_containing_fsel(sel_small)
        else:
            raise Exception("PointsSelection: 'sel_small' not PointsSelection or FieldSelection sel_small={sel_small}")

    def __repr__(self):
        """
        only show information of this node.
        """
        return f"PointsSelection({self.points_dist_type}, {self.total_site}, {self.xg_arr.tolist()})"

    def hash_sha256(self):
        """
        Return hash of psel.
        Always return the same hash from all the node.
        If points_dist_type == "g" AND the hash for all the nodes are the same, then return the hash for each node.
        Otherwise, return the hash of the gathered information from all the node.
        """
        v = q.hash_sha256((
            "PointsSelection:",
            self.xg_arr,
            self.total_site,
            self.points_dist_type,
            ))
        v_list = q.get_comm().allgather(v)
        if self.points_dist_type == "g":
            for v1 in v_list:
                if v != v1:
                    return q.hash_sha256(v_list)
            return v
        return q.hash_sha256(v_list)

    def __getstate__(self):
        """
        Only work if run with single node (or if all nodes has the same data).
        """
        xg_arr = self.xg_arr
        total_site = self.total_site
        points_dist_type = self.points_dist_type
        return [ xg_arr, total_site, points_dist_type, ]

    def __setstate__(self, state):
        """
        Only work if run with single node (or if all nodes has the same data).
        """
        if self.view_count > 0:
            raise ValueError("can't load while being viewed")
        [ xg_arr, total_site, points_dist_type, ] = state
        self.__init__(total_site, xg_arr, points_dist_type)

### -------------------------------------------------------------------

cdef class FieldSelection:

    def __cinit__(self):
        self.cdata = <cc.Long>&(self.xx)
        self.view_count = 0

    def __init__(self, *args):
        """
        FieldSelection()
        FieldSelection(geo, 0) # selecting all points
        FieldSelection(geo, -1) # no points being selected
        FieldSelection(geo) # same as `FieldSelection(geo, -1)`, no points being selected
        FieldSelection(psel) # require `psel.points_dist_type in [ "l", "f", "g", ]`
        FieldSelection(psel, geo) # require `psel.points_dist_type in [ "l", "f", "g", ]`
        """
        cdef cc.Int len_args = len(args)
        if len_args == 0:
            self.xx.init()
            return
        elif isinstance(args[0], Geometry):
            self.init_from_geo(*args)
        elif isinstance(args[0], PointsSelection):
            self.init_from_psel(*args)
        else:
            raise Exception("FieldSelection.__init__: {args}")

    def init_from_geo(self, Geometry geo, cc.Long rank=-1):
        """
        By default `rank=-1` means empty selection.
        Non-negative `rank` means full selection.
        """
        self.set_uniform(geo, rank)

    def init_from_psel(self, PointsSelection psel, Geometry geo=None, cc.Long rank_psel=1024 * 1024 * 1024 * 1024 * 1024):
        """
        psel.points_dist_type [ "g", "f", "l", ]
        """
        if geo is None:
            geo = psel.geo
        cc.set_fsel_from_psel(self.xx, psel.xx, geo.xx, rank_psel)

    def __getbuffer__(self, Py_buffer *buffer, int flags):
        """
        Get buffer view of fsel.f_rank field as 1-D array of np.int64
        The values are rank (rank >= 0 means selected, rank == -1 means not selected)
        Need to call fsel.update() after modifying the f_rank via this buffer view.
        """
        cdef int ndim = 1
        cdef Buffer buf = Buffer(self, ndim)
        buf.format = 'q'
        buf.itemsize = sizeof(cc.Int64t)
        buf.buf = <char*>(self.xx.f_rank.field.data())
        cdef int multiplicity = self.xx.f_rank.multiplicity
        assert multiplicity == 1
        buf.set_dim_size(0, self.xx.f_rank.field.size())
        buf.update_strides_from_shape()
        buf.set_buffer(buffer, flags)
        self.view_count += 1

    def release_buffer(self, Buffer buf):
        assert buf.obj is self
        self.view_count -= 1

    def __imatmul__(self, FieldSelection v1):
        cc.assign_direct(self.xx, v1.xx)
        return self

    def copy(self):
        x = type(self)()
        x @= self
        return x

    def __copy__(self):
        return self.copy()

    def __deepcopy__(self, memo):
        return self.copy()

    def update(self):
        """
        update various indices based on f_rank
        """
        cc.update_field_selection(self.xx)

    def set_empty(self, Geometry geo not None):
        """
        set an empty fsel with geo (all rank=-1)
        """
        if self.view_count > 0:
            raise Exception("FieldSelection: self.view_count>0")
        self.set_uniform(geo, -1)

    def set_uniform(self, Geometry geo not None, cc.Long val=0):
        """
        default (val = 0) select every sites
        val = -1 deselection everything
        """
        if self.view_count > 0:
            raise Exception("FieldSelection: self.view_count>0")
        self.xx.init()
        cc.mk_field_selection(self.xx.f_rank, geo.xx, val)
        self.update()

    def set_rand(self, Coordinate total_site not None, cc.Long n_per_tslice, RngState rs not None):
        if self.view_count > 0:
            raise Exception("FieldSelection: self.view_count>0")
        cc.mk_field_selection(self.xx.f_rank, total_site.xx, n_per_tslice, rs.xx)
        self.update()

    def set_rand_psel(self, Coordinate total_site not None, cc.Long n_per_tslice, RngState rs not None,
                      PointsSelection psel=None):
        if self.view_count > 0:
            raise Exception("FieldSelection: self.view_count>0")
        self.set_rand(total_site, n_per_tslice, rs)
        if psel is not None:
            self.add_psel(psel)
        self.update()

    def add_psel(self, PointsSelection psel, cc.Long rank_psel=1024 * 1024 * 1024 * 1024 * 1024):
        """
        Add psel points to the selection, with the rank specified as rank_psel.
        If the point is already selected with lower rank, the rank is unchanged.
        """
        cc.add_field_selection(self.xx.f_rank, psel.xx, rank_psel)
        self.update()

    def add_fsel(self, FieldSelection fsel):
        """
        Add fsel points to the selection, with the rank specified in fsel.
        If the point is already selected with lower rank, the rank is unchanged.
        """
        cc.add_field_selection(self.xx.f_rank, fsel.xx)
        self.update()

    def intersect_with(self, FieldSelection fsel):
        """
        Modify the `self`.
        More efficient if `self` is smaller than `fsel`.
        """
        cc.intersect_with(self.xx, fsel.xx)

    def intersect(self, FieldSelection fsel):
        """
        Do NOT change the `self`, but return a new one
        More efficient if `self` is smaller than `fsel`
        """
        cdef FieldSelection fsel_new = self.copy()
        fsel_new.intersect_with(fsel)
        return fsel_new

    def is_containing_psel(self, PointsSelection psel_small):
        cdef cc.Bool x = cc.is_containing(self.xx, psel_small.xx)
        return x

    def is_containing_fsel(self, FieldSelection fsel_small):
        cdef cc.Bool x = cc.is_containing(self.xx, fsel_small.xx)
        return x

    def is_containing(self, sel_small):
        if isinstance(sel_small, PointsSelection):
            return self.is_containing_psel(sel_small)
        elif isinstance(sel_small, FieldSelection):
            return self.is_containing_fsel(sel_small)
        else:
            raise Exception("PointsSelection: 'sel_small' not PointsSelection or FieldSelection sel_small={sel_small}")

    def to_psel(self):
        cdef PointsSelection psel = PointsSelection(self.total_site)
        cc.assign_direct(psel.xx, cc.psel_from_fsel(self.xx))
        return psel

    def to_psel_local(self):
        return PointsSelection(self)

    def save(self, const cc.std_string& path):
        cdef cc.Long total_bytes = cc.write_field_selection(self.xx, path)
        return total_bytes

    def load(self, const cc.std_string& path):
        if self.view_count > 0:
            raise Exception("FieldSelection: self.view_count>0")
        cdef cc.Long total_bytes = cc.read_field_selection(self.xx, path)
        return total_bytes

    @property
    def geo(self):
        cdef Geometry geo = Geometry()
        geo.xx = self.xx.get_geo()
        return geo

    @property
    def total_site(self):
        cdef Coordinate total_site = Coordinate()
        cc.assign_direct(total_site.xx, self.xx.get_geo().total_site())
        return total_site

    @property
    def n_elems(self):
        return self.xx.n_elems

    def __setitem__(self, idx, val):
        """
        Implemented in terms of ``np.asarray``
        """
        np.asarray(self)[idx] = val

    def __getitem__(self, idx):
        """
        Implemented in terms of ``np.asarray``
        """
        return np.asarray(self)[idx]

    def __iter__(self):
        """
        iterate over all local selected coordinate as xg
        """
        cdef cc.Long idx
        cdef cc.Long n_elems = self.n_elems
        for idx in range(n_elems):
            yield self.coordinate_from_idx(idx)

    def __len__(self):
        return self.n_elems

    def idx_from_coordinate(self, Coordinate xg not None):
        cdef cc.Coordinate xl_xx = self.xx.get_geo().coordinate_l_from_g(xg.xx)
        cdef cc.Long idx = self.xx.f_local_idx.get_elem(xl_xx)
        return idx

    def coordinate_from_idx(self, cc.Long idx):
        cdef Coordinate xg = Coordinate()
        cdef cc.Long index = self.xx.indices[idx]
        cdef cc.Coordinate xl_xx = self.xx.get_geo().coordinate_from_index(index)
        cc.assign_direct(xg.xx, self.xx.get_geo().coordinate_g_from_l(xl_xx))
        return xg

    def __getstate__(self):
        """
        Only work when single node.
        """
        geo = self.geo
        fsel_arr = self[:].copy()
        return [ fsel_arr, geo, ]

    def __setstate__(self, state):
        """
        Only work when single node.
        """
        if self.view_count > 0:
            raise ValueError("can't load while being viewed")
        self.__init__()
        cdef Geometry geo
        [ fsel_arr, geo, ] = state
        self.set_empty(geo)
        self[:] = fsel_arr
        self.update()

### -------------------------------------------------------------------

cache_points_selection = q.mk_cache("points_selection")

@q.timer
def mk_xg_field(Geometry geo):
    cdef FieldInt f = FieldInt()
    cc.set_xg_field(f.xx, geo.xx)
    return f

def get_psel_single(Coordinate total_site, Coordinate xg=None):
    """
    [ xg, ]
    """
    cdef PointsSelection psel
    if xg is None:
        xg = Coordinate([ -1, -1, -1, -1, ])
    param_tuple = (total_site[0], total_site[1], total_site[2], total_site[3], xg[0], xg[1], xg[2], xg[3],)
    if param_tuple not in cache_points_selection:
        psel = PointsSelection(total_site, xg,)
        cache_points_selection[param_tuple] = psel
    return cache_points_selection[param_tuple]

def get_psel_tslice(Coordinate total_site, *, int t_dir=3):
    """
    if t_dir = 3, then [ [-1,-1,-1,0,], [-1,-1,-1,1,], ..., [-1,-1,-1,total_site[3]-1], ]
    if t_dir = 2, then [ [-1,-1,0,-1,], [-1,-1,1,-1,], ..., [-1,-1,total_site[2]-1,-1], ]
    """
    cdef PointsSelection psel
    assert 0 <= t_dir and t_dir < 4
    param_tuple = (total_site[0], total_site[1], total_site[2], total_site[3], t_dir,)
    if param_tuple not in cache_points_selection:
        psel = PointsSelection(total_site)
        cc.assign_direct(psel.xx, cc.mk_tslice_points_selection(total_site.xx, t_dir))
        cache_points_selection[param_tuple] = psel
    return cache_points_selection[param_tuple]

def is_matching_fsel(FieldSelection fsel1, FieldSelection fsel2):
    return cc.is_matching_fsel(fsel1.xx, fsel2.xx)

### -------------------------------------------------------------------



import qlat as q
from . import rbc_ukqcd_params as rup
from .rbc_ukqcd_params import (
        set_param,
        get_param,
        get_job_seed,
        )

from qlat import is_test

import numpy as np

import functools
import os
from pprint import pformat

save_path_default = "results"

load_path_list = [ "results", ]

def get_save_path(fn):
    return f"{save_path_default}/{fn}"

def get_load_path(*fns):
    def get(fn):
        if fn is None:
            return None
        elif isinstance(fn, (tuple, list)):
            for f in fn:
                p = get(f)
                if p is not None:
                    return p
        else:
            for path in load_path_list:
                p = f"{path}/{fn}"
                if q.does_file_exist_qar_sync_node(p):
                    return p
        return None
    return get(fns)

# ----------

@q.timer_verbose
def check_job(job_tag, traj, fns_produce, fns_need):
    """
    return False if config is finished or unavailable
    """
    is_job_done = True
    for fn in fns_produce:
        if get_load_path(fn) is None:
            q.displayln_info(f"check_job: {job_tag} {traj} to do as '{fn}' does not exist.")
            is_job_done = False
            break
    if is_job_done:
        return False
    #
    is_job_avail = True
    for fn in fns_need:
        if get_load_path(fn) is None:
            q.displayln_info(f"check_job: {job_tag} {traj} unavailable as '{fn}' does not exist.")
            is_job_avail = False
            break
    if not is_job_avail:
        return False
    #
    q.check_stop()
    q.check_time_limit()
    #
    assert not is_job_done and is_job_avail
    #
    return True

# ----------

@q.timer_verbose
def run_params(job_tag):
    fname = q.get_fname()
    param = get_param(job_tag)
    q.json_results_append(q.json_dumps(param, indent=None))
    param_lines = q.json_dumps(param, indent=2).split("\n")
    for v in param_lines:
        q.displayln_info(f"CHECK: params: {job_tag}: {v}")
    path_dir = get_save_path(f"{job_tag}/params")
    fn_prefix = f"{path_dir}/version-"
    fn_suffix = ".json"
    fn_suffix2 = ".pickle"
    def mk_fn(version):
        path = f"{fn_prefix}{version:010}{fn_suffix}"
        return path
    fn_list = q.qls_sync_node(path_dir)
    assert isinstance(fn_list, list)
    if len(fn_list) == 0:
        version = 1
    else:
        version_list = []
        for fn in fn_list:
            assert fn.startswith(fn_prefix)
            assert fn.endswith(fn_suffix) or fn.endswith(fn_suffix2)
            if fn.endswith(fn_suffix):
                tag = fn
                tag = tag.removeprefix(fn_prefix)
                tag = tag.removesuffix(fn_suffix)
                assert len(tag) == 10
                v = int(tag)
                assert fn == mk_fn(v)
                version_list.append(v)
        version_list.sort()
        if len(version_list) == 0:
            version = 1
        else:
            version = version_list[-1]
            fn = mk_fn(version)
            fn_pickle = fn.removesuffix(fn_suffix) + fn_suffix2
            assert fn in fn_list
            assert fn_pickle in fn_list
            if param == q.load_pickle_obj(fn_pickle, is_sync_node=True):
                # assert param == q.load_json_obj(fn, is_sync_node=True) # not always true due to json format not very complete (dict keys cannot be integer).
                assert q.qcat_sync_node(fn) == q.json_dumps(param, indent=2)
                q.displayln_info(0, f"{fname}: loaded '{fn}' matches param exactly.")
                return
            version += 1
    fn = mk_fn(version)
    fn_pickle = fn.removesuffix(fn_suffix) + fn_suffix2
    q.save_pickle_obj(get_param(job_tag), fn_pickle, is_sync_node=True)
    q.save_json_obj(get_param(job_tag), fn, indent=2, is_sync_node=True)

# ----------

def mk_gf_fn_list(job_tag, traj):
    stream_skip = 1000000
    fn_list = [
            f"{job_tag}/configs/ckpoint_lat.{traj}",
            f"{job_tag}/configs/ckpoint_lat.IEEE64BIG.{traj}",
            ]
    if traj >= stream_skip:
        fn_list += [
                f"{job_tag}/configs-b/ckpoint_lat.{traj - stream_skip}",
                f"{job_tag}/configs-b/ckpoint_lat.IEEE64BIG.{traj - stream_skip}",
                ]
    if traj >= 2 * stream_skip:
        fn_list += [
                f"{job_tag}/configs-c/ckpoint_lat.{traj - 2 * stream_skip}",
                f"{job_tag}/configs-c/ckpoint_lat.IEEE64BIG.{traj - 2 * stream_skip}",
                ]
    if traj >= 3 * stream_skip:
        fn_list += [
                f"{job_tag}/configs-d/ckpoint_lat.{traj - 3 * stream_skip}",
                f"{job_tag}/configs-d/ckpoint_lat.IEEE64BIG.{traj - 3 * stream_skip}",
                ]
    if traj >= 4 * stream_skip:
        fn_list += [
                f"{job_tag}/configs-e/ckpoint_lat.{traj - 4 * stream_skip}",
                f"{job_tag}/configs-e/ckpoint_lat.IEEE64BIG.{traj - 4 * stream_skip}",
                ]
    return fn_list

@q.timer_verbose
def run_gf(job_tag, traj):
    path_gf = get_load_path(mk_gf_fn_list(job_tag, traj))
    if path_gf is None:
        if job_tag[:5] == "test-":
            if not q.obtain_lock(f"locks/{job_tag}-{traj}-gauge_field"):
                return None
            total_site = q.Coordinate(get_param(job_tag, "total_site"))
            gf = rup.mk_sample_gauge_field_v3(job_tag, f"{traj}")
            path_gf = get_save_path(f"{job_tag}/configs/ckpoint_lat.{traj}")
            gf.save(path_gf)
            q.release_lock()
        else:
            return None
    get_gf = rup.load_config_lazy(job_tag, path_gf)
    return get_gf

@q.timer_verbose
def run_gt(job_tag, traj, get_gf):
    tfn = f"{job_tag}/gauge-transform/traj-{traj}.field"
    path_gt = get_load_path(tfn)
    if path_gt is None:
        if None in [ get_gf, ]:
            return None
        elif q.obtain_lock(f"locks/{job_tag}-{traj}-gauge_fix_coulomb"):
            gf = get_gf()
            import qlat_gpt as qg
            gt = qg.gauge_fix_coulomb(gf)
            gt.save_cps(get_save_path(f"{job_tag}/gauge-transform/traj-{traj}.gfix"))
            gt.save_double(get_save_path(tfn))
            q.release_lock()
        else:
            return None
    @q.lazy_call
    @q.timer_verbose
    def load_gt():
        path_gt = get_load_path(tfn)
        assert path_gt is not None
        gt = q.GaugeTransform()
        gt.load_double(path_gt)
        # ADJUST ME
        # import qlat_gpt as qg
        # qg.check_gauge_fix_coulomb(get_gf(), gt)
        #
        return gt
    get_gt = load_gt
    return get_gt

# ----------

@q.timer
def mk_rand_wall_src_info_n_exact(job_tag, traj, inv_type):
    params = rup.dict_params[job_tag]
    n_exact = params["n_exact_wsrc"]
    seed = get_job_seed(job_tag)
    rs = q.RngState(f"seed {seed} {traj}").split("mk_rand_wall_src_info")
    inv_acc_s = 1
    inv_acc_e = 2
    total_site = q.Coordinate(get_param(job_tag, "total_site"))
    t_size = total_site[3]
    wi_s = [ [ t, inv_type, inv_acc_s, ] for t in range(t_size) ]
    mask = [ False, ] * t_size
    for i in range(n_exact):
        t_e = rs.rand_gen() % t_size
        mask[t_e] = True
    wi_e = []
    for t in range(t_size):
        if mask[t]:
            wi_e.append([ t, inv_type, inv_acc_e, ])
    wi = wi_e + wi_s
    for i in range(len(wi)):
        wi[i] = [ i, ] + wi[i]
    return wi

@q.timer
def mk_rand_wall_src_info_prob(job_tag, traj, inv_type):
    params = rup.dict_params[job_tag]
    prob = params["prob_exact_wsrc"]
    seed = get_job_seed(job_tag)
    rs = q.RngState(f"seed {seed} {traj}").split("mk_rand_wall_src_info_prob")
    inv_acc_s = 1
    inv_acc_e = 2
    total_site = q.Coordinate(get_param(job_tag, "total_site"))
    t_size = total_site[3]
    wi_s = [ [ t, inv_type, inv_acc_s, ] for t in range(t_size) ]
    wi_e = []
    for t in range(t_size):
        if rs.u_rand_gen() < prob:
            wi_e.append([ t, inv_type, inv_acc_e, ])
    wi = wi_e + wi_s
    for i in range(len(wi)):
        wi[i] = [ i, ] + wi[i]
    return wi

def get_prob_exact_wsrc(job_tag):
    params = rup.dict_params[job_tag]
    if "prob_exact_wsrc" in params:
        return params["prob_exact_wsrc"]
    n_exact = params["n_exact_wsrc"]
    total_site = q.Coordinate(get_param(job_tag, "total_site"))
    return 1 - (1 - 1 / total_site[3])**n_exact

@q.timer
def mk_rand_wall_src_info(job_tag, traj, inv_type):
    """
    wi is a list of [ idx tslice inv_type inv_acc ]
    """
    params = rup.dict_params[job_tag]
    if "prob_exact_wsrc" not in params:
        return mk_rand_wall_src_info_n_exact(job_tag, traj, inv_type)
    return mk_rand_wall_src_info_prob(job_tag, traj, inv_type)

@q.timer
def save_wall_src_info(wi, path):
    """
    wi is a list of  [ idx tslice inv_type inv_acc ]
    """
    if 0 != q.get_id_node():
        return None
    lines = [ " ".join([ f"{v:5d}" for v in l ]) for l in wi ]
    content = "\n".join(lines + [ "", ])
    q.qtouch(path, content)

@q.timer
def load_wall_src_info(path):
    assert path is not None
    """
    wi is a list of [ idx tslice inv_type inv_acc ]
    """
    dt = q.qload_datatable_sync_node(path, True)
    t = [ list(map(int, l)) for l in dt ]
    wi = [ [ l[0], l[1], l[2], l[3], ] for l in t ]
    return wi

@q.timer_verbose
def run_wi(job_tag, traj):
    tfn_l = f"{job_tag}/wall-src-info-light/traj-{traj}.txt"
    tfn_s = f"{job_tag}/wall-src-info-strange/traj-{traj}.txt"
    path_light = get_load_path(tfn_l)
    if path_light is None:
        if q.obtain_lock(f"locks/{job_tag}-{traj}-wi"):
            wi_light = mk_rand_wall_src_info(job_tag, traj, inv_type=0)
            save_wall_src_info(wi_light, get_save_path(tfn_l));
            q.release_lock()
        else:
            return None
    path_strange = get_load_path(tfn_s)
    if path_strange is None:
        if q.obtain_lock(f"locks/{job_tag}-{traj}-wi"):
            wi_strange = mk_rand_wall_src_info(job_tag, traj, inv_type=1)
            save_wall_src_info(wi_strange, get_save_path(tfn_s));
            q.release_lock()
        else:
            return None
    @q.lazy_call
    @q.timer_verbose
    def load():
        wi_light = load_wall_src_info(get_load_path(tfn_l))
        wi_strange = load_wall_src_info(get_load_path(tfn_s))
        return wi_light + wi_strange
    return load

# ----------

def get_n_points_psel(job_tag):
    assert job_tag in rup.dict_params
    total_site = q.Coordinate(get_param(job_tag, "total_site"))
    total_volume = total_site.volume()
    psel_rate = get_param(job_tag, "field_selection_psel_rate")
    if psel_rate is not None:
        n_points = round(total_volume * psel_rate)
        return n_points
    n_points = get_param(job_tag, "n_points_psel")
    assert n_points is not None
    return n_points

@q.timer
def mk_rand_psel(job_tag, traj):
    seed = get_job_seed(job_tag)
    rs = q.RngState(f"seed {seed} {traj}").split("mk_rand_psel")
    total_site = q.Coordinate(get_param(job_tag, "total_site"))
    n_points = get_n_points_psel(job_tag)
    psel = q.PointsSelection()
    psel.set_rand(total_site, n_points, rs)
    return psel

@q.timer_verbose
def run_psel(job_tag, traj):
    tfn = f"{job_tag}/points-selection/traj-{traj}.lati"
    path_psel = get_load_path(tfn)
    if path_psel is None:
        if q.obtain_lock(f"locks/{job_tag}-{traj}-psel"):
            psel = mk_rand_psel(job_tag, traj)
            psel.save(get_save_path(tfn))
            q.release_lock()
        else:
            return None
    #
    @q.lazy_call
    @q.timer_verbose
    def load_psel():
        path_psel = get_load_path(tfn)
        assert path_psel is not None
        total_site = q.Coordinate(get_param(job_tag, "total_site"))
        psel = q.PointsSelection()
        psel.load(path_psel, q.Geometry(total_site))
        assert psel.n_points == get_n_points_psel(job_tag)
        return psel
    return load_psel

# ----------

def get_n_points_pi(job_tag, traj, inv_type, inv_acc):
    assert job_tag in rup.dict_params
    assert "n_points" in rup.dict_params[job_tag]
    return rup.dict_params[job_tag]["n_points"][inv_type][inv_acc]

@q.timer
def mk_rand_point_src_info(job_tag, traj, psel):
    """
    pi is a list of [ idx xg inv_type inv_acc ]
    """
    seed = get_job_seed(job_tag)
    rs = q.RngState(f"seed {seed} {traj}").split("mk_rand_point_src_info")
    xg_list = psel.xg_arr().tolist()
    assert len(xg_list) == get_n_points_pi(job_tag, traj, 0, 0)
    g_pi = [ [] for _ in xg_list ]
    for inv_type in [ 0, 1, ]:
        for inv_acc in [ 0, 1, 2, ]:
            for i in range(get_n_points_pi(job_tag, traj, inv_type, inv_acc)):
                g_pi[i].append([ xg_list[i], inv_type, inv_acc ])
    pi = []
    for g in g_pi:
        pi += g
    for i in range(len(pi)):
        pi[i] = [ i, ] + pi[i]
    return pi

@q.timer
def save_point_src_info(pi, path):
    """
    pi is a list of [ idx xg inv_type inv_acc ]
    """
    if 0 != q.get_id_node():
        return None
    def mk_line(l):
        [ idx, xg, inv_type, inv_acc ] = l
        return f"{idx:5d}    {xg[0]:3d} {xg[1]:3d} {xg[2]:3d} {xg[3]:3d}    {inv_type:3d} {inv_acc:3d}"
    lines = list(map(mk_line, pi))
    content = "\n".join([ f"{len(lines)}" ] + lines + [ "" ])
    q.qtouch(path, content)

@q.timer
def load_point_src_info(path):
    """
    pi is a list of [ idx xg inv_type inv_acc ]
    """
    dt = q.qload_datatable_sync_node(path, True)
    t = [ list(map(int, l)) for l in dt ][1:]
    pi = [ [ l[0], l[1:5], l[5], l[6], ] for l in t ]
    return pi

@q.timer_verbose
def run_pi(job_tag, traj, get_psel):
    tfn = f"{job_tag}/point-src-info/traj-{traj}.txt"
    path = get_load_path(tfn)
    if path is None:
        if q.obtain_lock(f"locks/{job_tag}-{traj}-pi"):
            pi = mk_rand_point_src_info(job_tag, traj, get_psel())
            save_point_src_info(pi, get_save_path(tfn));
            q.release_lock()
        else:
            return None
    @q.lazy_call
    @q.timer_verbose
    def load():
        path = get_load_path(tfn)
        assert path is not None
        pi = load_point_src_info(path)
        return pi
    return load

# ----------

@q.timer
def load_point_distribution(job_tag):
    """
    return point_distribution
    where
    point_distribution[xg_rel] = probability of the relative coordinate of a point relative to a selected point equals to ``xg_rel``.
    xg_rel = (x, y, z, t,)
    x >= y >= z >= 0 and t >= 0
    n_points = get_n_points_psel(job_tag)
    """
    n_points = get_n_points_psel(job_tag)
    tfn = f"{job_tag}/point-distribution/point-distribution.txt"
    path = get_load_path(tfn)
    if path is None:
        return None
    dt = q.qload_datatable_sync_node(path, True)
    point_distribution = dict()
    for l in dt:
        x, y, z, t, prob = l
        x = int(x)
        y = int(y)
        z = int(z)
        t = int(t)
        point_distribution[(x, y, z, t,)] = prob * (n_points - 1) / n_points
    point_distribution[(0, 0, 0, 0,)] = 1.0 / n_points
    return point_distribution

def classify_rel_coordinate(xg_rel_arrary, total_site_array):
    """
    xg_rel_arrary = np.array(xg_rel)
    total_site_array = np.array(total_site)
    """
    total_site_half = total_site_array // 2
    xg_rel_arrary = xg_rel_arrary % total_site_array
    xg_rel_abs = total_site_half - abs(xg_rel_arrary - total_site_half)
    x, y, z, t = xg_rel_abs
    x, y, z = sorted([x, y, z])
    return (x, y, z, t,)

def get_point_xrel_prob(xg_rel_arrary, total_site_array, point_distribution, n_points):
    """
    xg_rel_arrary = np.array(xg_rel)
    total_site_array = np.array(total_site)
    point_distribution = load_point_distribution(job_tag)
    n_points = get_n_points_psel(job_tag)
    """
    if point_distribution is None:
        if np.all(xg_rel_arrary == 0):
            return 1.0 / n_points
        else:
            total_volume = np.prod(total_site_array)
            return (n_points - 1) / n_points / (total_volume - 1)
    xg_rel = classify_rel_coordinate(xg_rel_arrary, total_site_array)
    prob = point_distribution[xg_rel]
    return prob

# ----------

@q.timer
def mk_rand_fsel(job_tag, traj, n_per_tslice):
    seed = get_job_seed(job_tag)
    rs = q.RngState(f"seed {seed} {traj}").split("mk_rand_fsel")
    total_site = q.Coordinate(get_param(job_tag, "total_site"))
    fsel = q.FieldSelection()
    fsel.set_rand(total_site, n_per_tslice, rs)
    return fsel

@q.timer_verbose
def run_fsel(job_tag, traj):
    tfn = f"{job_tag}/field-selection/traj-{traj}.field"
    path_fsel = get_load_path(tfn)
    total_site = q.Coordinate(get_param(job_tag, "total_site"))
    fsel_rate = get_param(job_tag, "field_selection_fsel_rate", default=1/16)
    n_per_tslice = round(total_site[0] * total_site[1] * total_site[2] * fsel_rate)
    if path_fsel is None:
        if q.obtain_lock(f"locks/{job_tag}-{traj}-fsel"):
            fsel = mk_rand_fsel(job_tag, traj, n_per_tslice)
            fsel.save(get_save_path(tfn))
            q.release_lock()
            return lambda : fsel
        else:
            return None
    @q.lazy_call
    @q.timer_verbose
    def load_fsel():
        path_fsel = get_load_path(tfn)
        assert path_fsel is not None
        fsel = q.FieldSelection()
        total_size = fsel.load(path_fsel)
        assert total_size > 0
        return fsel
    return load_fsel

# ----------

@q.timer
def mk_fselc(fsel, psel):
    fname = q.get_fname()
    fselc = fsel.copy()
    if not fsel.is_containing(psel):
        q.displayln_info(f"WARNING: {fname}: fsel does not containing psel.")
        fselc.add_psel(psel)
    return fselc

@q.timer_verbose
def run_fselc(job_tag, traj, get_fsel, get_psel):
    if get_fsel is None:
        return None
    if get_psel is None:
        return None
    @q.lazy_call
    @q.timer_verbose
    def get():
        fselc = mk_fselc(get_fsel(), get_psel())
        return fselc
    return get

# ----------

@q.timer
def mk_rand_fsel_smear(job_tag, traj, n_per_tslice_smear):
    seed = get_job_seed(job_tag)
    rs = q.RngState(f"seed {seed} {traj}").split("mk_rand_fsel_smear")
    total_site = q.Coordinate(get_param(job_tag, "total_site"))
    fsel = q.FieldSelection()
    fsel.set_rand(total_site, n_per_tslice_smear, rs)
    return fsel

@q.timer_verbose
def run_psel_smear(job_tag, traj):
    """
    return lambda : psel_smear
    psel_smear should randomly select same number of point on each tslice
    """
    tfn = f"{job_tag}/points-selection-smear/traj-{traj}.lati"
    path_psel = get_load_path(tfn)
    total_site = q.Coordinate(get_param(job_tag, "total_site"))
    if path_psel is None:
        if q.obtain_lock(f"locks/{job_tag}-{traj}-psel-smear"):
            n_per_tslice_smear = get_param(job_tag, "n_per_tslice_smear")
            fsel = mk_rand_fsel_smear(job_tag, traj, n_per_tslice_smear)
            psel = fsel.to_psel()
            psel.save(get_save_path(tfn))
            q.release_lock()
        else:
            return None
    #
    @q.lazy_call
    @q.timer_verbose
    def load_psel():
        path_psel = get_load_path(tfn)
        assert path_psel is not None
        total_site = q.Coordinate(get_param(job_tag, "total_site"))
        psel = q.PointsSelection()
        psel.load(path_psel, q.Geometry(total_site))
        return psel
    return load_psel

@q.timer_verbose
def run_psel_smear_median(job_tag, traj):
    """
    return lambda : psel_smear_median
    psel_smear should randomly select same number of point on each tslice
    """
    fname = q.get_fname()
    tfn = f"{job_tag}/psel_smear_median/traj-{traj}.lati"
    path_psel = get_load_path(tfn)
    total_site = q.Coordinate(get_param(job_tag, "total_site"))
    if path_psel is None:
        if q.obtain_lock(f"locks/{job_tag}-{traj}-{fname}"):
            n_per_tslice_smear = get_param(job_tag, "n_per_tslice_smear_median")
            fsel = mk_rand_fsel_smear(job_tag, traj, n_per_tslice_smear)
            psel = fsel.to_psel()
            psel.save(get_save_path(tfn))
            q.release_lock()
        else:
            return None
    #
    @q.lazy_call
    @q.timer_verbose
    def load_psel():
        path_psel = get_load_path(tfn)
        assert path_psel is not None
        total_site = q.Coordinate(get_param(job_tag, "total_site"))
        psel = q.PointsSelection()
        psel.load(path_psel, q.Geometry(total_site))
        return psel
    return load_psel

# ----------

@q.timer_verbose
def run_gf_ape(job_tag, get_gf):
    if get_gf is None:
        return None
    coef = rup.dict_params[job_tag]["gf_ape_smear_coef"]
    step = rup.dict_params[job_tag]["gf_ape_smear_step"]
    #
    @q.lazy_call
    @q.timer_verbose
    def run():
        gf = get_gf()
        gf_ape = q.gf_spatial_ape_smear(gf, coef, step)
        gf_ape = q.mk_left_expanded_field(gf_ape)
        return gf_ape
    return run

@q.timer_verbose
def run_gf_hyp(job_tag, get_gf):
    if get_gf is None:
        return None
    step = get_param(job_tag, "gf_hyp_smear_step")
    @q.lazy_call
    @q.timer_verbose
    def run():
        gf_hyp = get_gf()
        for i in range(step):
            gf_hyp = q.gf_hyp_smear(gf_hyp, 0.75, 0.6, 0.3)
        return gf_hyp
    return run

# ----------

@q.timer_verbose
def compute_eig(job_tag, gf, inv_type=0, inv_acc=0, *, path=None, pc_ne=None):
    """
    return a function ``get_eig''
    ``get_eig()'' return the ``eig''
    """
    from . import rbc_ukqcd as ru
    load_eig = ru.load_eig_lazy(get_load_path(path), job_tag)
    if load_eig is not None:
        return load_eig
    import gpt as g
    # eig = ru.mk_eig(job_tag, gf, inv_type, inv_acc)
    eig = ru.mk_ceig(job_tag, gf, inv_type, inv_acc, pc_ne=pc_ne)
    ru.save_eig(get_save_path(path + ".partial"), eig, job_tag, inv_type, inv_acc);
    q.qrename_info(get_save_path(path + ".partial"), get_save_path(path))
    test_eig(job_tag, gf, eig, inv_type)
    def get_eig():
        return eig
    return get_eig

@q.timer_verbose
def test_eig(job_tag, gf, eig, inv_type, *, pc_ne=None):
    from .rbc_ukqcd import get_inv
    geo = gf.geo
    src = q.FermionField4d(geo)
    src.set_rand(q.RngState("test_eig:src.set_rand"))
    q.json_results_append(f"src norm", src.qnorm(), 1e-10)
    q.json_results_append(f"src get_data_sig", q.get_data_sig(src, q.RngState()), 1e-10)
    sol_ref = get_inv(gf, job_tag, inv_type, inv_acc=3, eig=eig, eps=1e-14, mpi_split=False, n_grouped=1, pc_ne=pc_ne, qtimer=False) * src
    q.json_results_append(f"sol_ref norm", sol_ref.qnorm(), 1e-10)
    q.json_results_append(f"sol_ref get_data_sig", q.get_data_sig(sol_ref, q.RngState()), 1e-10)
    for inv_acc in [ 0, 1, ]:
        sol = get_inv(gf, job_tag, inv_type, inv_acc, eig=eig, mpi_split=False, n_grouped=1, pc_ne=pc_ne, qtimer=False) * src
        sol -= sol_ref
        q.json_results_append(f"sol diff norm inv_acc={inv_acc} with eig", sol.qnorm(), 1e-2)
        q.json_results_append(f"sol diff get_data_sig inv_acc={inv_acc} with eig", q.get_data_sig(sol, q.RngState()), 1e-2)
        sol = get_inv(gf, job_tag, inv_type, inv_acc, mpi_split=False, n_grouped=1, pc_ne=pc_ne, qtimer=False) * src
        sol -= sol_ref
        q.json_results_append(f"sol diff norm inv_acc={inv_acc} without eig", sol.qnorm(), 1e-2)
        q.json_results_append(f"sol diff get_data_sig inv_acc={inv_acc} without eig", q.get_data_sig(sol, q.RngState()), 1e-2)

@q.timer_verbose
def run_eig(job_tag, traj, get_gf, *, is_only_load=False):
    if None in [ get_gf, ]:
        return None
    from . import rbc_ukqcd as ru
    get_eig = ru.load_eig_lazy(get_load_path(f"{job_tag}/eig/traj-{traj}"), job_tag)
    if is_only_load:
        return get_eig
    if get_eig is None and get_gf is not None:
        if q.obtain_lock(f"locks/{job_tag}-{traj}-run-eig"):
            get_eig = compute_eig(job_tag, get_gf(), inv_type=0, path=f"{job_tag}/eig/traj-{traj}")
            q.release_lock()
            return get_eig
        else:
            return None
    else:
        return get_eig

@q.timer_verbose
def run_eig_strange(job_tag, traj, get_gf, *, is_only_load=False):
    """
    if failed, return None
    if no parameter, return lambda : None
    """
    if None in [ get_gf, ]:
        return None
    if get_param(job_tag, "clanc_params", 1) is None:
        fn = f"{job_tag}/eig-strange/traj-{traj}/no-eig-parameters.txt"
        if get_load_path(fn) is None:
            q.qtouch_info(get_save_path(fn))
        return lambda : None
    from . import rbc_ukqcd as ru
    get_eig = ru.load_eig_lazy(get_load_path(f"{job_tag}/eig-strange/traj-{traj}"), job_tag)
    if is_only_load:
        return get_eig
    if get_eig is None and get_gf is not None:
        if q.obtain_lock(f"locks/{job_tag}-{traj}-run-eig-strange"):
            get_eig = compute_eig(job_tag, get_gf(), inv_type=1, path=f"{job_tag}/eig-strange/traj-{traj}")
            q.release_lock()
            return get_eig
        else:
            return None
    else:
        return get_eig

# ----------

@functools.lru_cache(maxsize=None)
def get_r_list(job_tag):
    total_site = q.Coordinate(get_param(job_tag, "total_site"))
    r_limit = q.get_r_limit(total_site)
    r_list = q.mk_r_list(r_limit, r_all_limit=28.0, r_scaling_factor=5.0)
    # r_list = q.mk_r_list(r_limit, r_all_limit=0.0, r_scaling_factor=5.0) # old choice
    return r_list

@functools.lru_cache(maxsize=None)
def get_r_sq_interp_idx_coef_list(job_tag):
    """
    Return [ (r_idx_low, r_idx_high, coef_low, coef_high,), ... ] indexed by r_sq
    """
    r_list = get_r_list(job_tag)
    return q.mk_r_sq_interp_idx_coef_list(r_list)

@q.timer_verbose
def run_r_list(job_tag):
    fn = f"{job_tag}/r_list/r_list.lat"
    r_list = get_r_list(job_tag)
    ld = q.mk_lat_data([
        [ "r_idx", len(r_list), ],
        ])
    ld.from_numpy(np.array(r_list))
    if get_load_path(fn) is not None:
        ld_load = q.load_lat_data(get_load_path(fn))
        assert ld.is_match(ld_load)
        ld_diff = ld - ld_load
        assert ld_diff.qnorm() < 1e-20
        return
    ld.save(get_save_path(fn))

# ----------



from qlat_utils import *
from .c import *
from . import c

@timer
def contract_chvp_16(prop1, prop2):
    """
    return chvp_16
    #
    inline void contract_chvp_16(
        FieldM<ComplexD, 16>& chvp_16,
        const Propagator4d& prop1_x_y,
        const Propagator4d& prop2_x_y)
    #
    chvp_16.get_elem(x, mu * 4 + nu) ==
    tr(g5_herm(prop2_x_y.get_elem(x)) * gammas[mu]
    * prop1_x_y.get_elem(x) * gammas[nu])
    #
    mu: polarization at sink location x
    nu: polarization at source location y
    """
    chvp_16 = Field(ElemTypeComplexD)
    c.contract_chvp_16_field(chvp_16, prop1, prop2)
    return chvp_16



# cython: binding=True, embedsignature=True, c_string_type=unicode, c_string_encoding=utf8

from qlat_utils.all cimport *
from . cimport everything as cc
from .geometry cimport Geometry
from .field_types cimport (
        FieldComplexD,
        )
from .field_selection cimport (
        FieldSelection,
        PointsSelection,
        )

from cpython cimport Py_buffer
from cpython.buffer cimport PyBUF_FORMAT

import cqlat as c
import qlat_utils as q
import numpy as np

from .field_utils import mk_fft

cdef class Prop(FieldWilsonMatrix):

    def __init__(self, Geometry geo=None, multiplicity=1):
        assert multiplicity == 1
        super().__init__(geo, 1)

    cdef cc.Handle[cc.Prop] xxx(self):
        assert self.xx.multiplicity == 1 or self.xx.multiplicity == 0
        return cc.Handle[cc.Prop](<cc.Prop&>self.xx)

    def get_elem_wm(self, cc.Long index, int m=0):
        cdef WilsonMatrix wm = WilsonMatrix()
        np.asarray(wm)[:] = self[index, m]
        return wm

    def __getstate__(self):
        """
        Only work when single node (or if all nodes has the same data).
        """
        return super().__getstate__()

    def __setstate__(self, state):
        """
        Only work when single node (or if all nodes has the same data).
        """
        super().__setstate__(state)

    def glb_sum_tslice(self, *, cc.Int t_dir=3):
        cdef SelectedPointsWilsonMatrix sp = super().glb_sum_tslice(t_dir=t_dir)
        cdef PselProp sp_prop = PselProp(sp.psel)
        sp_prop @= sp
        return sp_prop

###

cdef class SelProp(SelectedFieldWilsonMatrix):

    def __init__(self, FieldSelection fsel=None, multiplicity=1):
        assert multiplicity == 1
        super().__init__(fsel, 1)

    cdef cc.Handle[cc.SelProp] xxx(self):
        assert self.xx.multiplicity == 1 or self.xx.multiplicity == 0
        return cc.Handle[cc.SelProp](<cc.SelProp&>self.xx)

    def get_elem_wm(self, cc.Long idx, int m=0):
        cdef WilsonMatrix wm = WilsonMatrix()
        np.asarray(wm)[:] = self[idx, m]
        return wm

    def __getstate__(self):
        """
        Only work when single node (or if all nodes has the same data).
        """
        return super().__getstate__()

    def __setstate__(self, state):
        """
        Only work when single node (or if all nodes has the same data).
        """
        super().__setstate__(state)

###

cdef class PselProp(SelectedPointsWilsonMatrix):

    def __init__(self, *args):
        cdef cc.Int len_args = len(args)
        if len_args == 0:
            super().__init__()
        elif isinstance(args[0], PointsSelection):
            if len(args) == 1:
                psel, = args
            else:
                psel, multiplicity, = args
                assert multiplicity == 1
            super().__init__(psel, 1)
        elif isinstance(args[0], SelProp):
            super().__init__(*args)
        else:
            assert False

    cdef cc.Handle[cc.PselProp] xxx(self):
        assert self.xx.multiplicity == 1 or self.xx.multiplicity == 0
        return cc.Handle[cc.PselProp](<cc.PselProp&>self.xx)

    def get_elem_wm(self, cc.Long idx, int m=0):
        cdef WilsonMatrix wm = WilsonMatrix()
        np.asarray(wm)[:] = self[idx, m]
        return wm

    def __getstate__(self):
        """
        Only work when single node (or if all nodes has the same data).
        """
        return super().__getstate__()

    def __setstate__(self, state):
        """
        Only work when single node (or if all nodes has the same data).
        """
        super().__setstate__(state)

###

cdef class SpinProp(FieldSpinMatrix):

    def __init__(self, Geometry geo=None, multiplicity=1):
        assert multiplicity == 1
        super().__init__(geo, 1)

    cdef cc.Handle[cc.SpinProp] xxx(self):
        assert self.xx.multiplicity == 1 or self.xx.multiplicity == 0
        return cc.Handle[cc.SpinProp](<cc.SpinProp&>self.xx)

    def get_elem_sm(self, cc.Long index, int m=0):
        cdef SpinMatrix sm = SpinMatrix()
        np.asarray(sm)[:] = self[index, m]
        return sm

    def __getstate__(self):
        """
        Only work when single node (or if all nodes has the same data).
        """
        return super().__getstate__()

    def __setstate__(self, state):
        """
        Only work when single node (or if all nodes has the same data).
        """
        super().__setstate__(state)

    def glb_sum_tslice(self, *, cc.Int t_dir=3):
        cdef SelectedPointsSpinMatrix sp = super().glb_sum_tslice(t_dir=t_dir)
        return sp

###

def set_point_src(Prop prop_src not None, Geometry geo not None, Coordinate xg not None, cc.PyComplexD value=1.0):
    cc.set_point_src(prop_src.xxx().val(), geo.xx, xg.xx, cc.ccpy_d(value))

def set_wall_src(Prop prop_src not None, Geometry geo not None, int tslice, CoordinateD lmom=None):
    if lmom is None:
        lmom = CoordinateD()
    cc.set_wall_src(prop_src.xxx().val(), geo.xx, tslice, lmom.xx)

def set_rand_vol_u1(
        FieldComplexD fu1 not None,
        Geometry geo not None,
        cc.Int multiplicity,
        RngState rs not None,
        ):
    cc.set_rand_vol_u1(fu1.xx, geo.xx, multiplicity, rs.xx)

def set_rand_vol_u1_src(
        Prop prop_src not None,
        FieldComplexD fu1 not None,
        ):
    """
    prop_src ~ fu1
    """
    cc.set_rand_vol_u1_src(prop_src.xxx().val(), fu1.xx)

@q.timer
def mk_point_src(Geometry geo not None, Coordinate xg not None, cc.PyComplexD value=1.0):
    cdef Prop prop_src = Prop(geo)
    set_point_src(prop_src, geo, xg, value)
    return prop_src

@q.timer
def mk_wall_src(Geometry geo not None, int tslice, CoordinateD lmom=None):
    cdef Prop prop_src = Prop(geo)
    set_wall_src(prop_src, geo, tslice, lmom)
    return prop_src

@q.timer
def mk_rand_vol_u1(
        Geometry geo not None,
        cc.Int multiplicity,
        RngState rs not None,
        ):
    """
    return prop_src, fu1
    prop_src ~ fu1
    """
    cdef FieldComplexD fu1 = FieldComplexD(geo, multiplicity)
    set_rand_vol_u1(fu1, geo, multiplicity, rs)
    return fu1

@q.timer
def mk_rand_vol_u1_src(
        FieldComplexD fu1 not None,
        ):
    """
    return prop_src
    prop_src ~ fu1
    """
    cdef Prop prop_src = Prop(fu1.geo)
    set_rand_vol_u1_src(prop_src, fu1)
    return prop_src

@q.timer
def mk_rand_u1_src(sel, rs):
    """
    return (prop_src, fu1,) where prop_src = Prop() and fu1 = FieldComplex
    fu1 stores the random u1 numbers (fu1.multiplicity == 1)
    sel can be psel or fsel
    """
    prop_src = Prop()
    fu1 = FieldComplexD()
    if isinstance(sel, FieldSelection):
        fsel = sel
        c.set_rand_u1_src_fsel(prop_src, fu1, fsel, rs)
    elif isinstance(sel, PointsSelection):
        psel = sel
        geo = psel.geo
        assert isinstance(geo, Geometry)
        c.set_rand_u1_src_psel(prop_src, fu1, psel, geo, rs)
    else:
        raise Exception(f"mk_rand_u1_src {type(sel)}")
    return (prop_src, fu1,)

@q.timer
def get_rand_u1_sol(Prop prop_sol, FieldComplexD fu1, sel):
    assert isinstance(prop_sol, Prop)
    assert isinstance(fu1, FieldComplexD)
    if isinstance(sel, FieldSelection):
        fsel = sel
        s_prop = SelProp(fsel)
        c.set_rand_u1_sol_fsel(s_prop, prop_sol, fu1, fsel)
        return s_prop
    elif isinstance(sel, PointsSelection):
        psel = sel
        sp_prop = PselProp(psel)
        c.set_rand_u1_sol_psel(sp_prop, prop_sol, fu1, psel)
        return sp_prop
    else:
        raise Exception(f"get_rand_u1_sol {type(sel)}")

@q.timer_verbose
def mk_rand_u1_prop(inv, sel, rs):
    """
    interface function
    return s_prop
    sel can be psel or fsel
    """
    prop_src, fu1 = mk_rand_u1_src(sel, rs)
    prop_sol = inv * prop_src
    return get_rand_u1_sol(prop_sol, fu1, sel)

@q.timer
def free_invert(prop_src, cc.RealD mass, cc.RealD m5=1.0, CoordinateD momtwist=None):
    cdef Prop qcd_prop_src
    cdef Prop qcd_prop_sol
    cdef SpinProp spin_prop_src
    cdef SpinProp spin_prop_sol
    if momtwist is None:
        momtwist = CoordinateD([ 0.0, 0.0, 0.0, 0.0, ])
    if isinstance(prop_src, Prop):
        qcd_prop_src = prop_src
        qcd_prop_sol = Prop()
        cc.free_invert(qcd_prop_sol.xxx().val(), qcd_prop_src.xxx().val(), mass, m5, momtwist.xx)
        return qcd_prop_sol
    elif isinstance(prop_src, SpinProp):
        spin_prop_src = prop_src
        spin_prop_sol = SpinProp()
        cc.free_invert(spin_prop_sol.xxx().val(), spin_prop_src.xxx().val(), mass, m5, momtwist.xx)
        return spin_prop_sol
    else:
        assert False

@q.timer
def invert_qed(
        SpinProp sp_src, FieldComplexD gf1,
        cc.RealD mass, cc.RealD m5, cc.Int ls,
        *,
        t_wick_phase_factor_arr=None,
        cc.Bool is_dagger=False,
        cc.RealD stop_rsd=1e-8, cc.Long max_num_iter=50000,
        ):
    """
    gf1 = q.mk_left_expanded_field(gf)
    """
    cdef SpinProp sp_sol = SpinProp()
    cdef cc.vector[cc.ComplexD] t_wick_phase_factor_vec = cc.vector[cc.ComplexD]()
    cdef cc.Int t_size
    cdef cc.Int i
    if t_wick_phase_factor_arr is not None:
        t_size = len(t_wick_phase_factor_arr)
        t_wick_phase_factor_vec.resize(t_size)
        for i in range(t_size):
            t_wick_phase_factor_vec[i] = cc.ccpy_d(t_wick_phase_factor_arr[i])
    cc.invert_qed(
        sp_sol.xxx().val(), sp_src.xxx().val(), gf1.xx,
        mass, m5, ls, t_wick_phase_factor_vec,
        is_dagger, stop_rsd, max_num_iter)
    return sp_sol

def convert_mspincolor_from_wm(prop_wm):
    prop_msc = prop_wm.copy(False)
    if isinstance(prop_wm, Prop):
        c.convert_mspincolor_from_wm_prop(prop_msc, prop_wm)
    elif isinstance(prop_wm, SelProp):
        c.convert_mspincolor_from_wm_s_prop(prop_msc, prop_wm)
    elif isinstance(prop_wm, PselProp):
        c.convert_mspincolor_from_wm_sp_prop(prop_msc, prop_wm)
    else:
        raise Exception("prop type match failed")
    return prop_msc

def convert_wm_from_mspincolor(prop_msc):
    prop_wm = prop_msc.copy(False)
    if isinstance(prop_msc, Prop):
        c.convert_wm_from_mspincolor_prop(prop_wm, prop_msc)
    elif isinstance(prop_msc, SelProp):
        c.convert_wm_from_mspincolor_s_prop(prop_wm, prop_msc)
    elif isinstance(prop_msc, PselProp):
        c.convert_wm_from_mspincolor_sp_prop(prop_wm, prop_msc)
    else:
        raise Exception("prop type match failed")
    return prop_wm

@q.timer
def mk_ff_list_from_prop(Prop prop):
    """
    return ff_list
    isinstance(ff_list, list)
    len(ff_list) == 12
    """
    cdef cc.Int num_field = 12
    cdef list ff_list = []
    cdef FermionField4d ff
    cdef cc.std_vector[cc.FermionField4d] ff_vec
    cc.set_ff_vec_from_prop(ff_vec, prop.xxx().val())
    assert <cc.Int>ff_vec.size() == num_field
    for i in range(num_field):
        ff = FermionField4d()
        cc.qswap(ff_vec[i], ff.xx)
        ff_list.append(ff)
    return ff_list

@q.timer
def mk_prop_from_ff_list(list ff_list):
    """
    return prop
    isinstance(prop, Prop)
    """
    cdef cc.Int num_field = 12
    assert len(ff_list) == num_field
    cdef Prop prop = Prop()
    cdef FermionField4d ff
    cdef cc.std_vector[cc.FermionField4d] ff_vec
    ff_vec.resize(num_field)
    for i in range(num_field):
        ff = ff_list[i]
        cc.qswap(ff_vec[i], ff.xx)
    cc.set_prop_from_ff_vec(prop.xxx().val(), ff_vec)
    for i in range(num_field):
        ff = ff_list[i]
        cc.qswap(ff_vec[i], ff.xx)
    return prop

@q.timer
def flip_tpbc_with_tslice(prop, tslice_flip_tpbc):
    if isinstance(prop, SelProp):
        c.flip_tpbc_with_tslice_s_prop(prop, tslice_flip_tpbc)
    elif isinstance(prop, PselProp):
        c.flip_tpbc_with_tslice_sp_prop(prop, tslice_flip_tpbc)
    else:
        print(type(prop))
        assert False

@q.timer
def free_scalar_invert_mom_cfield(FieldComplexD f, mass):
    assert isinstance(f, FieldComplexD)
    c.free_scalar_invert_mom_cfield(f, mass)

@q.timer
def free_scalar_invert_cfield(src, mass, *, mode_fft=1):
    fft_f = mk_fft(is_forward=True, is_normalizing=True, mode_fft=mode_fft)
    fft_b = mk_fft(is_forward=False, is_normalizing=True, mode_fft=mode_fft)
    f = fft_f * src
    free_scalar_invert_mom_cfield(f, mass)
    sol = fft_b * f
    return sol

cdef class FermionField4d(FieldWilsonVector):

    def __init__(self, Geometry geo=None, multiplicity=1):
        assert multiplicity == 1
        super().__init__(geo, 1)

    cdef cc.Handle[cc.FermionField4d] xxx(self):
        assert self.xx.multiplicity == 1 or self.xx.multiplicity == 0
        return cc.Handle[cc.FermionField4d](<cc.FermionField4d&>self.xx)

###



from . cimport everything as cqlat

cdef class ShuffledFieldsWriter:

    cdef cqlat.ShuffledFieldsWriter xx
    cdef readonly cqlat.Long cdata

cdef class ShuffledFieldsReader:

    cdef cqlat.ShuffledFieldsReader xx
    cdef readonly set tags
    cdef readonly cqlat.Long cdata

cdef class ShuffledBitSet:

    cdef cqlat.ShuffledBitSet xx
    cdef readonly cqlat.Long cdata



import gpt as g
import numpy as np
import sys, os

g.message("Topological charge measurement with GPT")
g.message("by Christoph Lehner")
g.message("2024/05/17")

config = g.default.get_single("--config",None)

def smear(Uin, c1, epsilon):
    Usm = g.copy(Uin)
    if c1 == 0.0:
        action = g.qcd.gauge.action.wilson(2.0 * Uin[0].otype.shape[0])
    else:
        action = g.qcd.gauge.action.improved_with_rectangle(2.0 * Uin[0].otype.shape[0], c1)
    g.algorithms.integrator.euler(
        Usm, lambda: [g(-u) for u in action.gradient(Usm, Usm)], 1.0
    )(epsilon)
    return Usm

smear_info_list = [
    [ 20, 0.05, 0.0 ],
    [ 20, 0.05, 0.0 ],
    [ 20, 0.05, 0.0 ],
    [ 50, 0.01, -1.4008 ],
    [ 50, 0.01, -1.4008 ],
    [ 50, 0.01, -1.4008 ],
    [ 50, 0.01, -1.4008 ]
]

U = g.load(config)

vol3d = float(np.prod(U[0].grid.gdimensions[0:3]))

w = g.corr_io.writer(f"{config}.gluonic_luchang")
output = g.gpt_io.writer(f"{config}.gluonic_fields_luchang", mpi=[2,2,2,4])

# first plaquette
g.message("Plaquette")
P = g.slice(g.qcd.gauge.rectangle(U, 1, 1, field=True) / vol3d, 3)
w.write("P", P)

fQ = open(f"{config}.Q_luchang","wt")
fE = open(f"{config}.E_luchang","wt")

tau = 0.0
U_wf = U
c = {}
for nsmear, epsilon, c1 in smear_info_list:

    for i in range(nsmear):
        tau += epsilon
        U_wf = smear(U_wf, c1, epsilon)
        g.message("%g" % tau, g.qcd.gauge.plaquette(U_wf))

    g.message("Field Strength")
    E_field = g.qcd.gauge.energy_density(U_wf, field=True)
    E = g.slice(E_field / vol3d, 3)
    w.write("E(%g)" % tau, E)

    E = sum(E).real / len(E)
    t2E = tau**2 * E
    g.message("t2E = ", t2E)

    g.message("Topology")
    Q_field = g.qcd.gauge.topological_charge_5LI(U_wf, cache=c, field=True)
    Q = g.slice(Q_field / vol3d, 3)
    w.write("Q(%g)" % tau, Q)

    Q = sum(Q).real / len(Q)
    fQ.write("%g %.15g\n" % (tau, Q))
    fQ.flush()

    output.write({
        "t" : tau,
        "Q" : Q_field,
        "E" : E_field
    })
    output.flush()

    fE.write("%g %.15g\n" % (tau, E))
    fE.flush()

exit()



# Author: Luchang Jin 2022

import qlat as q
import sys

if len(sys.argv) < 2:
    q.displayln_info("Usage: fields-list path1 path2 ...")
    exit()

q.begin_with_mpi()

for path in sys.argv[1:]:
    q.displayln_info(0, path)
    sfr = q.open_fields(path, "r")
    tags = sfr.list()
    has_dup = sfr.has_duplicates()
    sfr.close()
    for tag in tags:
        q.displayln_info(tag)
    if has_dup:
        q.displayln_info(0, f"INFO: '{path}' has_duplicates.")

q.timer_display()

q.end_with_mpi()

exit()



# cython: binding=True, embedsignature=True, c_string_type=unicode, c_string_encoding=utf8

from qlat_utils.all cimport *
from . cimport everything as cc
from .field_types cimport (
    FieldInt,
    FieldRealD,
)
from .geometry cimport Geometry
from .qcd cimport (
    GaugeField,
    GaugeTransform,
)
from .hmc cimport GaugeMomentum
from .gauge_action cimport GaugeAction

import qlat_utils as q
from .hmc import gf_evolve
from .field_utils import (
    shuffle_field,
    shuffle_field_back,
)
from .mpi import (
    glb_sum,
)

@q.timer
def gf_energy_density(GaugeField gf):
    return cc.gf_energy_density(gf.xxx().val())

@q.timer
def gf_energy_density_field(GaugeField gf):
    cdef Geometry geo = gf.geo
    cdef FieldRealD fd = FieldRealD(geo, 1)
    cc.gf_energy_density_field(fd.xx, gf.xxx().val())
    return fd

@q.timer
def gf_energy_density_dir_field(GaugeField gf):
    r"""
    Similar to `gf_plaq_field`.
    For smeared gauge field:
    `gf_energy_density_dir_field(gf)[:]` \approx `6 * (1 - gf_plaq_field(gf)[:])`
    `gf_energy_density_dir_field(gf)[:].sum(-1, keepdims=True) == gf_energy_density_field(gf)[:]`
    See
    https://arxiv.org/pdf/1006.4518.pdf Eq. (2.1) (Fig. 1) (approximate Eq. (3.1))
    https://arxiv.org/pdf/1203.4469.pdf
    """
    cdef Geometry geo = gf.geo
    cdef FieldRealD fd = FieldRealD(geo, 6)
    cc.gf_energy_density_dir_field(fd.xx, gf.xxx().val())
    return fd

@q.timer
def gf_plaq_flow_force(GaugeField gf, FieldRealD plaq_factor):
    """
    Compute force with plaq dependent beta factor (relative to standard
    ``set_wilson_flow_z``).
    ``plaq_factor.multiplicity == 6``.
    Check ``gf_plaq_field`` for the order of plaq.
    """
    cdef GaugeMomentum gm_force = GaugeMomentum()
    cc.set_plaq_flow_z(gm_force.xxx().val(), gf.xxx().val(), plaq_factor.xx)
    return gm_force

# ----------------------

@q.timer
def gf_wilson_flow_force(GaugeField gf, cc.RealD c1=0.0):
    cdef GaugeMomentum gm_force = GaugeMomentum()
    cc.set_wilson_flow_z(gm_force.xxx().val(), gf.xxx().val(), c1)
    return gm_force

@q.timer
def gf_wilson_flow_step(GaugeField gf, cc.RealD epsilon, *, cc.RealD c1=0.0, str wilson_flow_integrator_type=None):
    """
    Modify `gf` in place.
    default: Runge-Kutta scheme
    http://arxiv.org/abs/1006.4518v3
    """
    if wilson_flow_integrator_type is None:
        wilson_flow_integrator_type = "runge-kutta"
    if wilson_flow_integrator_type == "runge-kutta":
        cc.gf_wilson_flow_step(gf.xxx().val(), epsilon, c1=c1)
    elif wilson_flow_integrator_type == "euler":
        cc.gf_wilson_flow_step_euler(gf.xxx().val(), epsilon, c1=c1)

@q.timer
def gf_energy_derivative_density_field(GaugeField gf, *, cc.RealD epsilon=0.0125, cc.RealD c1=0.0, str wilson_flow_integrator_type=None):
    cdef GaugeField gf1 = gf.copy()
    gf_wilson_flow_step(gf1, epsilon, c1=c1, wilson_flow_integrator_type=wilson_flow_integrator_type)
    cdef FieldRealD fd1 = gf_energy_density_field(gf1)
    gf1 @= gf
    gf_wilson_flow_step(gf1, -epsilon, c1=c1, wilson_flow_integrator_type=wilson_flow_integrator_type)
    cdef FieldRealD fd2 = gf_energy_density_field(gf1)
    fd1 -= fd2
    fd1 *= 1 / (2 * epsilon)
    return fd1

@q.timer
def gf_wilson_flow(GaugeField gf, cc.RealD flow_time, cc.Long steps,
        *, cc.RealD c1=0.0, cc.RealD existing_flow_time=0.0, str wilson_flow_integrator_type=None):
    fname = q.get_fname()
    epsilon = flow_time / steps
    energy_density_list = []
    for i in range(steps):
        gf_wilson_flow_step(gf, epsilon, c1=c1, wilson_flow_integrator_type=wilson_flow_integrator_type)
        t = (i + 1) * epsilon + existing_flow_time
        energy_density = gf_energy_density(gf)
        energy_density_list.append(energy_density)
        q.displayln_info(f"{fname}: t={t} ; E={energy_density} ; t^2 E={t*t*energy_density}")
    return energy_density_list

@q.timer
def gf_block_stout_smear(GaugeField gf, Coordinate block_site, cc.RealD step_size):
    """
    Stout smear in place.
    If block_site = q.Coordinate(), then no blocking.
    Otherwise, perform stout smearing for each block separately.
    """
    cc.gf_block_stout_smear(gf.xxx().val(), gf.xxx().val(), block_site.xx, step_size)

@q.timer
def gf_local_stout_smear(GaugeField gf, Coordinate block_site, cc.RealD step_size):
    """
    Stout smear in place.
    Otherwise, perform stout smearing for local field only.
    No communiation involved.
    If block_site = q.Coordinate(), then no blocking.
    Otherwise, perform stout smearing for each block separately.
    """
    cc.gf_local_stout_smear(gf.xxx().val(), gf.xxx().val(), block_site.xx, step_size)

@q.timer
def gf_local_avg_plaq(GaugeField gf, Coordinate block_site):
    """
    Compuate local (no glb_sum) average of plaq for plaq within with each blocks.
    Used for diagnostic in `gt_local_tree_gauge`.
    If `block_site == q.Coordinate()`, then `block_site = geo.node_site`.
    Otherwise, perform stout smearing for each block separately.
    """
    return cc.gf_local_avg_plaq(gf.xxx().val(), block_site.xx)

@q.timer
def mk_local_tree_gauge_f_dir(
        Geometry geo,
        Coordinate block_site,
        cc.Bool is_uniform,
        RngState rs,
        ):
    """
    return f_dir
    Used for `gt_local_tree_gauge(gf, f_dir)`
    if dir == 5 for a lattice site, then the point is one of the tree root,
    i.e. if f_dir[x] == 5, then gt[x] is identity.
    """
    cdef FieldInt f_dir = FieldInt()
    cc.set_local_tree_gauge_f_dir(f_dir.xx, geo.xx, block_site.xx, is_uniform, rs.xx)
    return f_dir

@q.timer
def gt_local_tree_gauge(GaugeField gf, FieldInt f_dir, cc.Int num_step):
    """
    return gt_inv
    Note that the gauge fixed gauge field should be obtained as
    gf_gt = gt_inv.inv() * gf
    This is opposite to the other gauge tranformation functions.
    is_uniform = True
    block_site = q.Coordinate([ 4, 4, 4, 4, ])
    f_dir = mk_local_tree_gauge_f_dir(gf.geo, block_site, is_uniform, rs)
    """
    cdef Geometry geo = gf.geo
    cdef GaugeTransform gt_inv = GaugeTransform(geo)
    cc.gt_local_tree_gauge(gt_inv.xxx().val(), gf.xxx().val(), f_dir.xx, num_step)
    return gt_inv

@q.timer
def gt_block_tree_gauge(
        GaugeField gf,
        *,
        Coordinate block_site=None,
        Coordinate new_size_node=None,
        cc.Bool is_uniform=True,
        cc.RealD stout_smear_step_size=0.125,
        cc.Int num_smear_step=4,
        list f_dir_list=None,
        RngState rs_f_dir=None,
        ):
    """
    return (gt_inv, f_dir_list,)
    Provide `f_dir_list`, otherwise `rs_f_dir` will be used to generate `f_dir_list`.
    If `f_dir_list` is provided, `rs_f_dir` will be ignored.
    Default `block_site = Coordinate([ 4, 4, 4, 4, ])`
    Default `rs_f_dir = RngState("seed-gt_block_tree_gauge-rs_f_dir")`
    """
    fname = q.get_fname()
    if block_site is None:
        block_site = Coordinate([ 4, 4, 4, 4, ])
    num_step = 0
    for mu in range(4):
        num_step += block_site[mu] // 2
    cdef Geometry geo = gf.geo
    cdef Coordinate total_site = geo.total_site
    if new_size_node is None:
        new_size_node = total_site // block_site
    cdef Coordinate new_node_site = total_site // new_size_node
    assert new_node_site * new_size_node == total_site
    assert (new_node_site // block_site) * block_site == new_node_site
    cdef Coordinate size_node = geo.size_node
    cdef cc.Bool is_shuffle = size_node != new_size_node
    cdef list gf_list
    if is_shuffle:
        gf_list = shuffle_field(gf, new_size_node)
    else:
        gf_list = [ gf.copy(), ]
    cdef list geo_list = [ gf_local.geo for gf_local in gf_list ]
    if f_dir_list is None:
        if rs_f_dir is None:
            rs_f_dir = RngState("seed-gt_block_tree_gauge-rs_f_dir")
        f_dir_list = [
            mk_local_tree_gauge_f_dir(geo_local, block_site, is_uniform, rs_f_dir)
            for geo_local in geo_list
            ]
    cdef list gt_inv_list = []
    cdef GaugeTransform gt_inv
    avg_plaq_sum = 0.0
    avg_plaq_count = 0
    for gf_local, f_dir_local in zip(gf_list, f_dir_list):
        for step in range(num_smear_step):
            gf_local_stout_smear(gf_local, block_site, stout_smear_step_size)
        avg_plaq_count += 1
        avg_plaq_sum = gf_local_avg_plaq(gf_local, block_site)
        gt_inv = gt_local_tree_gauge(gf_local, f_dir_local, num_step)
        gt_inv_list.append(gt_inv)
    gt_inv = GaugeTransform(geo)
    if is_shuffle:
        shuffle_field_back(gt_inv, gt_inv_list, new_size_node)
    else:
        [ gt_inv, ] = gt_inv_list
    avg_plaq = glb_sum(avg_plaq_sum) / glb_sum(avg_plaq_count)
    q.displayln_info(0, f"{fname}: avg_plaq = {avg_plaq}")
    return (gt_inv, f_dir_list,)


# cython: binding=True, embedsignature=True, c_string_type=unicode, c_string_encoding=utf8

import os

os.environ['CUBACORES'] = '0'

from . cimport everything as cc
from qlat_utils.all cimport *
cimport numpy
import numpy as np
import qlat_utils as q

@q.timer
def has_cuba():
    cdef cc.Bool b = cc.has_cuba()
    return b

@q.timer
def test_integration_multi_dimensional():
    cc.test_integrationMultidimensional()

@q.timer
def clear_muon_line_interpolations():
    """
    interface to C++ function `clear_muon_line_interpolations`
    """
    cc.clear_muon_line_interpolations()

def get_number_of_muon_line_interpolations():
    """
    interface to C++ function `get_number_of_muon_line_interpolations`
    """
    return cc.get_number_of_muon_line_interpolations()

@q.timer
def compute_save_muonline_interpolation(cc.std_string path, cc.std_vector[cc.Int] dims, object eps):
    """
    compute and save the muonline interpolation at `path`.
    #
    interface to C++ function `compute_save_muonline_interpolation_cc`
    #
    default: epsabs=1e-8 ; epsrel=1e-3
    dims = [ 6, 6, 6, 6, 6, ] ~ [ 16, 16, 16, 16, 16, ]
    """
    cdef cc.IntegrationEps cc_eps
    epsabs, epsrel, mineval, maxeval = eps
    cc_eps.epsabs = epsabs
    cc_eps.epsrel = epsrel
    cc_eps.mineval = mineval
    cc_eps.maxeval = maxeval
    return cc.compute_save_muonline_interpolation_cc(path, dims, cc_eps)

@q.timer
def load_multiple_muonline_interpolations(cc.std_string path, cc.std_vector[cc.Long] idx_list):
    """
    load muonline interpolation saved at `path`.
    #
    interface to C++ function `load_multiple_muonline_interpolations`
    #
    May load multiple interpolations if under path we have:
    f"{path}/{idx:010d}/checkpoint" where f"{path}/{idx:010d}" is the path for `idx` range from 0 up to some `limit`.
    Load all if idx_list is empty.
    """
    return cc.load_multiple_muonline_interpolations(path, idx_list)

def get_muon_line_m_extra_weights():
    """
    return weights as list of list of float.
    """
    return cc.get_muon_line_m_extra_weights()

def set_muon_line_m_extra_weights(object weights=None):
    """
    set weights as list of list of float or None (use default value).
    """
    if weights is None:
        weights = cc.get_muon_line_m_extra_weights_default()
    cc.set_muon_line_m_extra_weights(weights)

@q.timer
def calc_muon_line_m(CoordinateD x, CoordinateD y, object eps):
    """
    return ret
    ret is 4D np.array with total len=3*4*4*4 and dtype=np.float64
    ret[i, rho, sigma, lambda] = \mathcal M_{i,\\rho,\\sigma,\\lambda}(x,y,z)
    See Eq. (9) of https://arxiv.org/pdf/2304.04423.pdf
    #
    interface to C++ function `muon_line_sym_py`
    default: epsabs=1e-8 ; epsrel=1e-3
    """
    cdef cc.IntegrationEps cc_eps
    epsabs, epsrel, mineval, maxeval = eps
    cc_eps.epsabs = epsabs
    cc_eps.epsrel = epsrel
    cc_eps.mineval = mineval
    cc_eps.maxeval = maxeval
    cdef cc.std_vector[cc.RealD] vec = cc.muon_line_sym_py(x.xx, y.xx, cc_eps)
    cdef numpy.ndarray arr = np.ascontiguousarray(vec, dtype=np.float64).reshape(3, 4, 4, 4)
    return arr

def get_muon_line_m(CoordinateD x, CoordinateD y, CoordinateD z, cc.Int idx, object eps):
    """
    return ret
    ret is 4D np.array with total len=3*4*4*4 and dtype=np.float64
    ret[i, rho, sigma, lambda] = \mathcal M_{i,\\rho,\\sigma,\\lambda}(x,y,z)
    See Eq. (9) of https://arxiv.org/pdf/2304.04423.pdf
    #
    interface to C++ function `get_muon_line_m_py`
    if idx < 0: calculate instead of using loaded interpolation.
    else using the muonline interpolation loaded as `idx` (epsabs, epsrel, mineval, maxeval are ignored).
    """
    cdef cc.IntegrationEps cc_eps
    epsabs, epsrel, mineval, maxeval = eps
    cc_eps.epsabs = epsabs
    cc_eps.epsrel = epsrel
    cc_eps.mineval = mineval
    cc_eps.maxeval = maxeval
    cdef cc.std_vector[cc.RealD] vec = cc.get_muon_line_m_py(x.xx, y.xx, z.xx, idx, cc_eps)
    cdef numpy.ndarray arr = np.ascontiguousarray(vec, dtype=np.float64).reshape(3, 4, 4, 4)
    return arr

def get_muon_line_m_extra(CoordinateD x, CoordinateD y, CoordinateD z, cc.Int tag):
    """
    return ret
    ret is 4D np.array with total len=3*4*4*4 and dtype=np.float64
    ret[i, rho, sigma, lambda] = \mathcal M_{i,\\rho,\\sigma,\\lambda}(x,y,z)
    See Eq. (9) of https://arxiv.org/pdf/2304.04423.pdf
    #
    interface to C++ function `get_muon_line_m_extra_py`
    tag = 0 sub
    tag = 1 nosub
    #
    Interpolations should be loaded for this function:
    0: 6^5 with-sub
    1: 8^5 with-sub (used for tag 0)
    2: 10^5 with-sub
    3: 12^5 with-sub (used for tag 0)
    4: 14^5 with-sub
    5: 16^5 with-sub (used for tag 0)
    6: 6^5 no-sub
    7: 8^5 no-sub (used for tag 1)
    8: 10^5 no-sub
    9: 12^5 no-sub (used for tag 1)
    l0: 14^5 no-sub
    11: 16^5 no-sub (used for tag 1)
    """
    cdef cc.std_vector[cc.RealD] vec = cc.get_muon_line_m_extra_py(x.xx, y.xx, z.xx, tag)
    cdef numpy.ndarray arr = np.ascontiguousarray(vec, dtype=np.float64).reshape(3, 4, 4, 4)
    return arr

def get_muon_line_m_extra_lat(Coordinate x, Coordinate y, Coordinate z, Coordinate total_site, cc.RealD a, cc.Int tag):
    """
    return ret
    ret is 4D np.array with total len=3*4*4*4 and dtype=np.float64
    ret[i, rho, sigma, lambda] = \mathcal M_{i,\\rho,\\sigma,\\lambda}(x,y,z)
    See Eq. (9) of https://arxiv.org/pdf/2304.04423.pdf
    #
    interface to C++ function `get_muon_line_m_extra_lat_py`
    a is muon_mass in lattice unit.
    tag = 0 sub
    tag = 1 nosub
    #
    Interpolations should be loaded for this function:
    0: 6^5 with-sub
    1: 8^5 with-sub (used for tag 0)
    2: 10^5 with-sub
    3: 12^5 with-sub (used for tag 0)
    4: 14^5 with-sub
    5: 16^5 with-sub (used for tag 0)
    6: 6^5 no-sub
    7: 8^5 no-sub (used for tag 1)
    8: 10^5 no-sub
    9: 12^5 no-sub (used for tag 1)
    l0: 14^5 no-sub
    11: 16^5 no-sub (used for tag 1)
    """
    cdef cc.std_vector[cc.RealD] vec = cc.get_muon_line_m_extra_lat_py(x.xx, y.xx, z.xx, total_site.xx, a, tag)
    cdef numpy.ndarray arr = np.ascontiguousarray(vec, dtype=np.float64).reshape(3, 4, 4, 4)
    return arr



import sys
import qlat as q
import numpy as np
from pprint import pformat

size_node_list = [
        [1, 1, 1, 1],
        [1, 1, 1, 2],
        [1, 1, 1, 3],
        [1, 1, 1, 4],
        [1, 1, 1, 6],
        [1, 1, 1, 8],
        [1, 2, 2, 4],
        [2, 2, 2, 4],
        [2, 2, 2, 4],
        ]

q.begin_with_mpi(size_node_list)

if len(sys.argv) == 1:
    q.displayln_info("Usage: topo-measure-wilson-flow [ --source source_config ] [ --output path-for-results ] [ --density-field ] [ --flow-time 6 ] [ --flow-n-step-per-unit-time 80 ]")

q.displayln_info("Topological charge measurement based on Wilson flow with Qlattice")
q.displayln_info("by Luchang Jin")
q.displayln_info("2024/05/20")

p_source = q.get_arg("--source")
p_output = q.get_arg("--output")
p_flow_time = q.get_arg("--flow-time", "6")
p_flow_n_step = q.get_arg("--flow-n-step-per-unit-time", "80")
is_density_field = q.get_option("--density-field")

flow_time = int(p_flow_time)
flow_n_step = int(p_flow_n_step)

info_path = p_output

if is_density_field:
    assert p_output is not None
    density_field_path=info_path
else:
    density_field_path=None

smear_info_list = [
        [ 1.0 / flow_n_step, flow_n_step, 0.0, "runge-kutta", ],
        ] * flow_time

energy_derivative_info = [ 1.0 / flow_n_step, 0.0, "runge-kutta", ]

def load():
    if p_source is None:
        q.displayln_info("Need to provide source file with '--source filename'. Use a sample gauge field for now.")
        total_site = q.Coordinate([ 4, 4, 4, 8, ])
        geo = q.Geometry(total_site)
        gf = q.GaugeField(geo)
        rs = q.RngState("seed")
        gf.set_rand(rs.split("gf-init"), 0.5, 10)
    else:
        gf = q.GaugeField()
        gf.load(p_source)
    q.clear_all_caches()
    q.clear_mem_cache()
    return gf

gf = load()

topo_list, energy_list, = q.smear_measure_topo(
        gf,
        smear_info_list=smear_info_list,
        energy_derivative_info=energy_derivative_info,
        info_path=info_path,
        density_field_path=density_field_path,
        )

if info_path is None:
    q.displayln_info("To save the result, use '--output path'. Print to screen for now.")
    q.displayln_info(pformat(topo_list))
    q.displayln_info(pformat(energy_list))

q.timer_display()

q.end_with_mpi()

exit()
